"""
A component that multiplies a vector by 2, where the
size of the vector is given as an option of type 'int'.
"""
import numpy as np

import openmdao.api as om


class VectorDoublingComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('size', types=int)

    def setup(self):
        size = self.options['size']

        self.add_input('x', shape=size)
        self.add_output('y', shape=size)

    def setup_partials(self):
        size = self.options['size']
        self.declare_partials('y', 'x', val=2.,
                              rows=np.arange(size),
                              cols=np.arange(size))

    def compute(self, inputs, outputs):
        outputs['y'] = 2 * inputs['x']



import unittest
import numpy as np

from openmdao.api import ExplicitComponent, Problem, IndepVarComp, slicer

from openmdao.utils.array_utils import evenly_distrib_idxs
from openmdao.utils.mpi import MPI

try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None

from openmdao.utils.assert_utils import assert_near_equal


class DistributedAdder(ExplicitComponent):
    """
    Distributes the work of adding 10 to every item in the param vector
    """

    def initialize(self):
        self.options.declare('local_size', types=int, default=1,
                             desc="Local size of input and output vectors.")

    def setup(self):
        """
        specify the local sizes of the variables and which specific indices this specific
        distributed component will handle. Indices do NOT need to be sequential or
        contiguous!
        """
        local_size = self.options['local_size']
        self.add_input('x', val=np.zeros(local_size, float), distributed=True)
        self.add_output('y', val=np.zeros(local_size, float), distributed=True)

    def compute(self, inputs, outputs):

        # NOTE: Each process will get just its local part of the vector
        # print('process {0:d}: {1}'.format(self.comm.rank, params['x'].shape))

        outputs['y'] = inputs['x'] + 10.


class Summer(ExplicitComponent):
    """
    Aggregation component that collects all the values from the distributed
    vector addition and computes a total
    """

    def initialize(self):
        self.options.declare('size', types=int, default=1,
                             desc="Size of input and output vectors.")

    def setup(self):
        # NOTE: this component depends on the full y array, so OpenMDAO
        #       will automatically gather all the values for it
        self.add_input('y', val=np.zeros(self.options['size']))
        self.add_output('sum', 0.0, shape=1)

    def compute(self, inputs, outputs):
        outputs['sum'] = np.sum(inputs['y'])


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class DistributedAdderTest(unittest.TestCase):

    N_PROCS = 3

    def test_distributed_adder(self):
        size = 100  # how many items in the array

        prob = Problem()

        # NOTE: evenly_distrib_idxs is a helper function to split the array
        #       up as evenly as possible
        comm = prob.comm
        rank = comm.rank
        sizes, offsets = evenly_distrib_idxs(comm.size, size)
        local_size, local_offset = sizes[rank], offsets[rank]

        start = local_offset
        end = local_offset + local_size

        prob.model.add_subsystem('des_vars', IndepVarComp('x', np.ones(size)), promotes=['x'])
        prob.model.add_subsystem('plus', DistributedAdder(local_size=int(local_size)), promotes_outputs=['y'])
        summer = prob.model.add_subsystem('summer', Summer(size=size), promotes_outputs=['sum'])

        prob.model.promotes('plus', inputs=['x'], src_indices=np.arange(start, end, dtype=int))
        prob.model.promotes('summer', inputs=['y'], src_indices=slicer[:])

        prob.setup()

        prob['x'] = np.ones(size)

        prob.run_driver()

        inp = summer._inputs['y']
        for i in range(size):
            diff = 11.0 - inp[i]
            if diff > 1.e-6 or diff < -1.e-6:
                raise RuntimeError("Summer input y[%d] is %f but should be 11.0" %
                                   (i, inp[i]))

        assert_near_equal(prob['sum'], 11.0 * size, 1.e-6)


if __name__ == "__main__":
    from openmdao.utils.mpi import mpirun_tests
    mpirun_tests()



""" Test the KSFunction component. """
import unittest

import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.simple_comps import DoubleArrayComp
from openmdao.test_suite.test_examples.beam_optimization.multipoint_beam_stress import MultipointBeamGroup
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import force_check_partials


class TestKSFunction(unittest.TestCase):

    def test_basic_ks(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('px', om.IndepVarComp(name="x", val=np.ones((2, ))))
        model.add_subsystem('comp', DoubleArrayComp())
        model.add_subsystem('ks', om.KSComp(width=2))
        model.connect('px.x', 'comp.x1')
        model.connect('comp.y2', 'ks.g')

        model.add_design_var('px.x')
        model.add_objective('comp.y1')
        model.add_constraint('ks.KS', upper=0.0)

        prob.setup()
        prob.run_driver()

        assert_near_equal(max(prob['comp.y2']), prob['ks.KS'][0])

    def test_bad_units(self):
        with self.assertRaises(ValueError) as ctx:
            om.KSComp(units='wtfu')
        self.assertEqual(str(ctx.exception), "The units 'wtfu' are invalid.")

    def test_vectorized(self):
        prob = om.Problem()
        model = prob.model

        x = np.zeros((3, 5))
        x[0, :] = np.array([3.0, 5.0, 11.0, 13.0, 17.0])
        x[1, :] = np.array([13.0, 11.0, 5.0, 17.0, 3.0])*2
        x[2, :] = np.array([11.0, 3.0, 17.0, 5.0, 13.0])*3

        model.add_subsystem('ks', om.KSComp(width=5, vec_size=3))

        model.add_design_var('ks.g')
        model.add_constraint('ks.KS', upper=0.0)

        prob.setup()
        prob.set_val('ks.g', x)
        prob.run_driver()


        assert_near_equal(prob.get_val('ks.KS', indices=0), 17.0)
        assert_near_equal(prob.get_val('ks.KS', indices=1), 34.0)
        assert_near_equal(prob.get_val('ks.KS', indices=2), 51.0)

        partials = force_check_partials(prob, includes=['ks'], out_stream=None)

        for (of, wrt) in partials['ks']:
            assert_near_equal(partials['ks'][of, wrt]['abs error'][0], 0.0, 1e-6)

    def test_partials_no_compute(self):
        prob = om.Problem()

        model = prob.model

        model.add_subsystem('px', om.IndepVarComp('x', val=np.array([5.0, 4.0])))

        ks_comp = model.add_subsystem('ks', om.KSComp(width=2))

        model.connect('px.x', 'ks.g')

        prob.setup()
        prob.run_driver()

        # compute partials with the current model inputs
        inputs = { 'g': prob['ks.g'] }
        partials = {}

        ks_comp.compute_partials(inputs, partials)
        assert_near_equal(partials[('KS', 'g')], np.array([1., 0.]), 1e-6)

        # swap inputs and call compute partials again, without calling compute
        inputs['g'][0][0] = 4
        inputs['g'][0][1] = 5

        ks_comp.compute_partials(inputs, partials)
        assert_near_equal(partials[('KS', 'g')], np.array([0., 1.]), 1e-6)

    def test_beam_stress(self):
        E = 1.
        L = 1.
        b = 0.1
        volume = 0.01
        max_bending = 100.0

        num_cp = 5
        num_elements = 25
        num_load_cases = 2

        prob = om.Problem(model=MultipointBeamGroup(E=E, L=L, b=b, volume=volume, max_bending = max_bending,
                                                    num_elements=num_elements, num_cp=num_cp,
                                                    num_load_cases=num_load_cases))

        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        prob.driver.options['tol'] = 1e-9
        prob.driver.options['disp'] = False

        prob.setup(mode='rev')

        prob.run_driver()

        stress0 = prob['parallel.sub_0.stress_comp.stress_0']
        stress1 = prob['parallel.sub_0.stress_comp.stress_1']

        # Test that the maximum constraint prior to aggregation is close to "active".
        assert_near_equal(max(stress0), 100.0, tolerance=5e-2)
        assert_near_equal(max(stress1), 100.0, tolerance=5e-2)

        # Test that no original constraint is violated.
        self.assertTrue(np.all(stress0 < 100.0))
        self.assertTrue(np.all(stress1 < 100.0))

    def test_beam_stress_ks_add_constraint(self):
        E = 1.
        L = 1.
        b = 0.1
        volume = 0.01
        max_bending = 100.0

        num_cp = 5
        num_elements = 25
        num_load_cases = 2

        prob = om.Problem(model=MultipointBeamGroup(E=E, L=L, b=b, volume=volume, max_bending = max_bending,
                                                    num_elements=num_elements, num_cp=num_cp,
                                                    num_load_cases=num_load_cases,
                                                    ks_add_constraint=True))

        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        prob.driver.options['tol'] = 1e-9
        prob.driver.options['disp'] = False

        prob.setup(mode='rev')

        prob.run_driver()

        stress0 = prob['parallel.sub_0.stress_comp.stress_0']
        stress1 = prob['parallel.sub_0.stress_comp.stress_1']

        # Test that the the maximum constraint prior to aggregation is close to "active".
        assert_near_equal(max(stress0), 100.0, tolerance=5e-2)
        assert_near_equal(max(stress1), 100.0, tolerance=5e-2)

        # Test that no original constraint is violated.
        self.assertTrue(np.all(stress0 < 100.0))
        self.assertTrue(np.all(stress1 < 100.0))

    def test_upper(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y = 3.0*x', x=np.zeros((2, )), y=np.zeros((2, ))))
        model.add_subsystem('ks', om.KSComp(width=2))

        model.connect('comp.y', 'ks.g')

        model.ks.options['upper'] = 16.0
        prob.setup()
        prob.set_val('comp.x', np.array([5.0, 4.0]))
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS', indices=0), -1.0)

    def test_lower_flag(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y = 3.0*x', x=np.zeros((2, )), y=np.zeros((2, ))))

        model.add_subsystem('ks', om.KSComp(width=2))

        model.connect('comp.y', 'ks.g')

        model.ks.options['lower_flag'] = True
        prob.setup()
        prob.set_val('comp.x', np.array([5.0, 4.0]))
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS', indices=0), -12.0)


class TestKSFunctionFeatures(unittest.TestCase):

    def test_basic(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y = 3.0*x',
                                                x=np.zeros((2, )),
                                                y=np.zeros((2, ))), promotes_inputs=['x'])

        model.add_subsystem('ks', om.KSComp(width=2))

        model.connect('comp.y', 'ks.g')

        prob.setup()
        prob.set_val('x', np.array([5.0, 4.0]))
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS'), [[15.0]])

    def test_vectorized(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y = 3.0*x',
                                                x=np.zeros((2, 2)),
                                                y=np.zeros((2, 2))), promotes_inputs=['x'])
        model.add_subsystem('ks', om.KSComp(width=2, vec_size=2))

        model.connect('comp.y', 'ks.g')

        prob.setup()
        prob.set_val('x', np.array([[5.0, 4.0], [10.0, 8.0]]))
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS'), np.array([[15], [30]]))

    def test_upper(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y = 3.0*x',
                                                x=np.zeros((2, )),
                                                y=np.zeros((2, ))), promotes_inputs=['x'])
        model.add_subsystem('ks', om.KSComp(width=2))

        model.connect('comp.y', 'ks.g')

        model.ks.options['upper'] = 16.0
        prob.setup()
        prob.set_val('x', np.array([5.0, 4.0]))
        prob.run_model()

        assert_near_equal(prob['ks.KS'], np.array([[-1.0]]))

    def test_lower_flag(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y = 3.0*x',
                                                x=np.zeros((2, )),
                                                y=np.zeros((2, ))), promotes_inputs=['x'])

        model.add_subsystem('ks', om.KSComp(width=2))

        model.connect('comp.y', 'ks.g')

        model.ks.options['lower_flag'] = True
        prob.setup()
        prob.set_val('x', np.array([5.0, 4.0]))
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS'), [[-12.0]])

    def test_add_constraint(self):

        n = 50
        prob = om.Problem()
        model = prob.model

        prob.driver = om.ScipyOptimizeDriver()

        model.add_subsystem('comp', om.ExecComp('y = -3.0*x**2 + k',
                                                x=np.zeros((n, )),
                                                y=np.zeros((n, )),
                                                k=0.0), promotes_inputs=['x', 'k'])

        model.add_subsystem('ks', om.KSComp(width=n, upper=4.0, add_constraint=True))

        model.add_design_var('k', lower=-10, upper=10)
        model.add_objective('k', scaler=-1)

        model.connect('comp.y', 'ks.g')

        prob.setup()
        prob.set_val('x', np.linspace(-np.pi/2, np.pi/2, n))
        prob.set_val('k', 5.)

        prob.run_driver()

        self.assertTrue(max(prob.get_val('comp.y')) <= 4.0)

    def test_units(self):

        n = 10

        model = om.Group()

        model.add_subsystem('ks', om.KSComp(width=n, units='m'), promotes_inputs=[('g', 'x')])
        model.set_input_defaults('x', range(n), units='ft')

        prob = om.Problem(model=model)
        prob.setup()
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS', indices=0), np.amax(prob.get_val('x')), tolerance=1e-8)

    def test_minimum(self):

        n = 10

        model = om.Group()

        model.add_subsystem('ks', om.KSComp(width=n, minimum=True), promotes_inputs=[('g', 'x')])
        model.set_input_defaults('x', range(n))

        prob = om.Problem(model=model)
        prob.setup()
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS', indices=0), np.amin(prob.get_val('x')), tolerance=1e-8)

    def test_minimum_partials(self):

        n = 10

        model = om.Group()

        model.add_subsystem('ks', om.KSComp(width=n, minimum=True), promotes_inputs=[('g', 'x')])
        model.set_input_defaults('x', range(n))

        prob = om.Problem(model=model)
        prob.setup(force_alloc_complex=True)
        prob.run_model()

        partials = force_check_partials(prob, includes=['ks'], out_stream=None, method="cs", step=1e-200)
        assert_check_partials(partials)

    def test_minimum_and_lower_flag(self):

        n = 10

        model = om.Group()

        model.add_subsystem('ks', om.KSComp(width=n, minimum=True, lower_flag=True), promotes_inputs=[('g', 'x')])
        model.set_input_defaults('x', range(n))

        prob = om.Problem(model=model)
        prob.setup()
        prob.run_model()

        assert_near_equal(prob.get_val('ks.KS', indices=0), -np.amax(prob.get_val('x')), tolerance=1e-8)

    def test_minimum_and_lower_flag_partials(self):

        n = 10

        model = om.Group()

        model.add_subsystem('ks', om.KSComp(width=n, minimum=True, lower_flag=True), promotes_inputs=[('g', 'x')])
        model.set_input_defaults('x', range(n))

        prob = om.Problem(model=model)
        prob.setup(force_alloc_complex=True)
        prob.run_model()

        partials = force_check_partials(prob, includes=['ks'], out_stream=None, method="cs", step=1e-200)
        assert_check_partials(partials)

if __name__ == "__main__":
    unittest.main()



"""Define the ExplicitComponent class."""
import inspect

import numpy as np
from types import MethodType


from openmdao.jacobians.dictionary_jacobian import DictionaryJacobian
from openmdao.core.component import Component
from openmdao.vectors.vector import _full_slice
from openmdao.utils.class_util import overrides_method
from openmdao.recorders.recording_iteration_stack import Recording
from openmdao.core.constants import INT_DTYPE, _UNDEFINED
from openmdao.utils.jax_utils import jax, jit, ExplicitCompJaxify, \
    compute_partials as _jax_compute_partials, \
    compute_jacvec_product as _jax_compute_jacvec_product, ReturnChecker, _jax_register_pytree_class


_tuplist = (tuple, list)


class ExplicitComponent(Component):
    """
    Class to inherit from when all output variables are explicit.

    Parameters
    ----------
    **kwargs : dict of keyword arguments
        Keyword arguments that will be mapped into the Component options.

    Attributes
    ----------
    _has_compute_partials : bool
        If True, the instance overrides compute_partials.
    _vjp_hash : int or None
        Hash value for the last set of inputs to the compute_primal function.
    _vjp_fun : function or None
        The vector-Jacobian product function.
    """

    def __init__(self, **kwargs):
        """
        Store some bound methods so we can detect runtime overrides.
        """
        super().__init__(**kwargs)

        self._has_compute_partials = overrides_method('compute_partials', self, ExplicitComponent)
        self.options.undeclare('assembled_jac_type')
        self._vjp_hash = None
        self._vjp_fun = None

    @property
    def nonlinear_solver(self):
        """
        Get the nonlinear solver for this system.
        """
        return self._nonlinear_solver

    @nonlinear_solver.setter
    def nonlinear_solver(self, solver):
        """
        Raise an exception.
        """
        raise RuntimeError(f"{self.msginfo}: Explicit components don't support nonlinear solvers.")

    @property
    def linear_solver(self):
        """
        Get the linear solver for this system.
        """
        return self._linear_solver

    @linear_solver.setter
    def linear_solver(self, solver):
        """
        Raise an exception.
        """
        raise RuntimeError(f"{self.msginfo}: Explicit components don't support linear solvers.")

    def _configure(self):
        """
        Configure this system to assign children settings and detect if matrix_free.
        """
        if self.matrix_free is _UNDEFINED:
            self.matrix_free = overrides_method('compute_jacvec_product', self, ExplicitComponent)

    def _jac_wrt_iter(self, wrt_matches=None):
        """
        Iterate over (name, start, end, vec, slice, dist_sizes) for each column var in the jacobian.

        Parameters
        ----------
        wrt_matches : set or None
            Only include row vars that are contained in this set.  This will determine what
            the actual offsets are, i.e. the offsets will be into a reduced jacobian
            containing only the matching columns.

        Yields
        ------
        str
            Absolute name of 'wrt' variable.
        int
            Starting index.
        int
            Ending index.
        Vector
            The _inputs vector.
        slice
            A full slice.
        ndarray or None
            Distributed sizes if var is distributed else None
        """
        start = end = 0
        local_ins = self._var_abs2meta['input']
        toidx = self._var_allprocs_abs2idx
        sizes = self._var_sizes['input']
        for wrt, meta in self._var_abs2meta['input'].items():
            if wrt_matches is None or wrt in wrt_matches:
                end += meta['size']
                vec = self._inputs if wrt in local_ins else None
                dist_sizes = sizes[:, toidx[wrt]] if meta['distributed'] else None
                yield wrt, start, end, vec, _full_slice, dist_sizes
                start = end

    def _setup_residuals(self):
        """
        Prevent the user from implementing setup_residuals for explicit components.
        """
        if overrides_method('setup_residuals', self, ExplicitComponent):
            raise RuntimeError(f'{self.msginfo}: Class overrides setup_residuals but '
                               'is an ExplicitComponent. setup_residuals may only be '
                               'overridden by ImplicitComponents.')

    def _setup_partials(self):
        """
        Call setup_partials in components.
        """
        super()._setup_partials()

        if self.matrix_free:
            return

        # Note: These declare calls are outside of setup_partials so that users do not have to
        # call the super version of setup_partials. This is still in the final setup.
        for out_abs, meta in self._var_abs2meta['output'].items():

            size = meta['size']
            if size > 0:

                # ExplicitComponent jacobians have -1 on the diagonal.
                arange = np.arange(size, dtype=INT_DTYPE)

                self._subjacs_info[out_abs, out_abs] = {
                    'rows': arange,
                    'cols': arange,
                    'shape': (size, size),
                    'val': np.full(size, -1.),
                    'dependent': True,
                }

    def _setup_jacobians(self, recurse=True):
        """
        Set and populate jacobian.

        Parameters
        ----------
        recurse : bool
            If True, setup jacobians in all descendants. (ignored)
        """
        if self._has_approx and self._use_derivatives:
            self._set_approx_partials_meta()

    def add_output(self, name, val=1.0, shape=None, units=None, res_units=None, desc='',
                   lower=None, upper=None, ref=1.0, ref0=0.0, res_ref=None, tags=None,
                   shape_by_conn=False, copy_shape=None, compute_shape=None, distributed=None):
        """
        Add an output variable to the component.

        For ExplicitComponent, res_ref defaults to the value in res unless otherwise specified.

        Parameters
        ----------
        name : str
            Name of the variable in this component's namespace.
        val : float or list or tuple or ndarray
            The initial value of the variable being added in user-defined units. Default is 1.0.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        res_units : str or None
            Units in which the residuals of this output will be given to the user when requested.
            Default is None, which means it has no units.
        desc : str
            Description of the variable.
        lower : float or list or tuple or ndarray or None
            Lower bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no lower bound.
            Default is None.
        upper : float or list or tuple or ndarray or None
            Upper bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no upper bound.
            Default is None.
        ref : float
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 1. Default is 1.
        ref0 : float
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 0. Default is 0.
        res_ref : float
            Scaling parameter. The value in the user-defined res_units of this output's residual
            when the scaled value is 1. Default is None, which means residual scaling matches
            output scaling.
        tags : str or list of strs
            User defined tags that can be used to filter what gets listed when calling
            list_inputs and list_outputs and also when listing results from case recorders.
        shape_by_conn : bool
            If True, shape this output to match its connected input(s).
        copy_shape : str or None
            If a str, that str is the name of a variable. Shape this output to match that of
            the named variable.
        compute_shape : function or None
            If a function, that function is called to determine the shape of this output.
        distributed : bool
            If True, this variable is a distributed variable, so it can have different sizes/values
            across MPI processes.

        Returns
        -------
        dict
            Metadata for added variable.
        """
        if res_ref is None:
            res_ref = ref

        return super().add_output(name, val=val, shape=shape, units=units,
                                  res_units=res_units, desc=desc,
                                  lower=lower, upper=upper,
                                  ref=ref, ref0=ref0, res_ref=res_ref,
                                  tags=tags, shape_by_conn=shape_by_conn,
                                  copy_shape=copy_shape, compute_shape=compute_shape,
                                  distributed=distributed)

    def _approx_subjac_keys_iter(self):
        is_output = self._outputs._contains_abs
        for abs_key, meta in self._subjacs_info.items():
            if 'method' in meta and not is_output(abs_key[1]):
                method = meta['method']
                if (method is not None and method in self._approx_schemes):
                    yield abs_key

    def _compute_wrapper(self):
        """
        Call compute based on the value of the "run_root_only" option.
        """
        with self._call_user_function('compute'):
            if self._run_root_only():
                if self.comm.rank == 0:
                    if self._discrete_inputs or self._discrete_outputs:
                        self.compute(self._inputs, self._outputs,
                                     self._discrete_inputs, self._discrete_outputs)
                    else:
                        self.compute(self._inputs, self._outputs)
                    self.comm.bcast([self._outputs.asarray(), self._discrete_outputs], root=0)
                else:
                    new_outs, new_disc_outs = self.comm.bcast(None, root=0)
                    self._outputs.set_val(new_outs)
                    if new_disc_outs:
                        for name, val in new_disc_outs.items():
                            self._discrete_outputs[name] = val
            else:
                if self._discrete_inputs or self._discrete_outputs:
                    self.compute(self._inputs, self._outputs,
                                 self._discrete_inputs, self._discrete_outputs)
                else:
                    self.compute(self._inputs, self._outputs)

    def _apply_nonlinear(self):
        """
        Compute residuals. The model is assumed to be in a scaled state.
        """
        outputs = self._outputs
        residuals = self._residuals
        with self._unscaled_context(outputs=[outputs], residuals=[residuals]):
            residuals.set_vec(outputs)

            # Sign of the residual is minus the sign of the output vector.
            residuals *= -1.0
            self._compute_wrapper()
            residuals += outputs
            outputs -= residuals

        self.iter_count_apply += 1

    def _solve_nonlinear(self):
        """
        Compute outputs. The model is assumed to be in a scaled state.
        """
        with Recording(self.pathname + '._solve_nonlinear', self.iter_count, self):
            with self._unscaled_context(outputs=[self._outputs], residuals=[self._residuals]):
                self._residuals.set_val(0.0)
                self._compute_wrapper()

            # Iteration counter is incremented in the Recording context manager at exit.

    def _compute_jacvec_product_wrapper(self, inputs, d_inputs, d_resids, mode,
                                        discrete_inputs=None):
        """
        Call compute_jacvec_product based on the value of the "run_root_only" option.

        Parameters
        ----------
        inputs : Vector
            Nonlinear input vector.
        d_inputs : Vector
            Linear input vector.
        d_resids : Vector
            Linear residual vector.
        mode : str
            Indicates direction of derivative computation, either 'fwd' or 'rev'.
        discrete_inputs : dict or None
            Mapping of variable name to discrete value.
        """
        if self._run_root_only():
            if self.comm.rank == 0:
                if discrete_inputs:
                    self.compute_jacvec_product(inputs, d_inputs, d_resids, mode, discrete_inputs)
                else:
                    self.compute_jacvec_product(inputs, d_inputs, d_resids, mode)
                if mode == 'fwd':
                    self.comm.bcast(d_resids.asarray(), root=0)
                else:  # rev
                    self.comm.bcast(d_inputs.asarray(), root=0)
            else:
                new_vals = self.comm.bcast(None, root=0)
                if mode == 'fwd':
                    d_resids.set_val(new_vals)
                else:  # rev
                    d_inputs.set_val(new_vals)
        else:
            dochk = mode == 'rev' and self._problem_meta['checking'] and self.comm.size > 1

            if dochk:
                nzdresids = self._get_dist_nz_dresids()

            if discrete_inputs:
                self.compute_jacvec_product(inputs, d_inputs, d_resids, mode, discrete_inputs)
            else:
                self.compute_jacvec_product(inputs, d_inputs, d_resids, mode)

            if dochk:
                self._check_consistent_serial_dinputs(nzdresids)

    def _apply_linear(self, jac, mode, scope_out=None, scope_in=None):
        """
        Compute jac-vec product. The model is assumed to be in a scaled state.

        Parameters
        ----------
        jac : Jacobian or None
            If None, use local jacobian, else use jac.
        mode : str
            'fwd' or 'rev'.
        scope_out : set or None
            Set of absolute output names in the scope of this mat-vec product.
            If None, all are in the scope.
        scope_in : set or None
            Set of absolute input names in the scope of this mat-vec product.
            If None, all are in the scope.
        """
        J = self._jacobian if jac is None else jac

        with self._matvec_context(scope_out, scope_in, mode) as vecs:
            d_inputs, d_outputs, d_residuals = vecs

            if not self.matrix_free:
                # if we're not matrix free, we can skip the rest because
                # compute_jacvec_product does nothing.

                # Jacobian and vectors are all scaled, unitless
                J._apply(self, d_inputs, d_outputs, d_residuals, mode)
                return

            # Jacobian and vectors are all unscaled, dimensional
            with self._unscaled_context(outputs=[self._outputs], residuals=[d_residuals]):

                # set appropriate vectors to read_only to help prevent user error
                if mode == 'fwd':
                    d_inputs.read_only = True
                else:  # rev
                    d_residuals.read_only = True

                try:
                    # handle identity subjacs (output_or_resid wrt itself)
                    if J is None or isinstance(J, DictionaryJacobian):
                        if d_outputs._names:
                            rflat = d_residuals._abs_get_val
                            oflat = d_outputs._abs_get_val
                            subjacs_empty = len(self._subjacs_info) == 0

                            # 'val' in the code below is a reference to the part of the
                            # output or residual array corresponding to the variable 'v'
                            if mode == 'fwd':
                                for v in d_outputs._names:
                                    if subjacs_empty or (v, v) not in self._subjacs_info:
                                        val = rflat(v)
                                        val -= oflat(v)
                            else:  # rev
                                for v in d_outputs._names:
                                    if subjacs_empty or (v, v) not in self._subjacs_info:
                                        val = oflat(v)
                                        val -= rflat(v)

                    # We used to negate the residual here, and then re-negate after the hook
                    with self._call_user_function('compute_jacvec_product'):
                        self._compute_jacvec_product_wrapper(self._inputs, d_inputs, d_residuals,
                                                             mode, self._discrete_inputs)
                finally:
                    d_inputs.read_only = d_residuals.read_only = False

    def _solve_linear(self, mode, scope_out=_UNDEFINED, scope_in=_UNDEFINED):
        """
        Apply inverse jac product. The model is assumed to be in a scaled state.

        Parameters
        ----------
        mode : str
            'fwd' or 'rev'.
        scope_out : set, None, or _UNDEFINED
            Outputs relevant to possible lower level calls to _apply_linear on Components.
        scope_in : set, None, or _UNDEFINED
            Inputs relevant to possible lower level calls to _apply_linear on Components.
        """
        d_outputs = self._doutputs
        d_residuals = self._dresiduals

        if mode == 'fwd':
            if self._has_resid_scaling:
                with self._unscaled_context(outputs=[d_outputs], residuals=[d_residuals]):
                    d_outputs.set_vec(d_residuals)
            else:
                d_outputs.set_vec(d_residuals)

            # ExplicitComponent jacobian defined with -1 on diagonal.
            d_outputs *= -1.0

        else:  # rev
            if self._has_resid_scaling:
                with self._unscaled_context(outputs=[d_outputs], residuals=[d_residuals]):
                    d_residuals.set_vec(d_outputs)
            else:
                d_residuals.set_vec(d_outputs)

            # ExplicitComponent jacobian defined with -1 on diagonal.
            d_residuals *= -1.0

    def _compute_partials_wrapper(self):
        """
        Call compute_partials based on the value of the "run_root_only" option.
        """
        with self._call_user_function('compute_partials'):
            if self._run_root_only():
                if self.comm.rank == 0:
                    if self._discrete_inputs:
                        self.compute_partials(self._inputs, self._jacobian, self._discrete_inputs)
                    else:
                        self.compute_partials(self._inputs, self._jacobian)
                    self.comm.bcast(list(self._jacobian.items()), root=0)
                else:
                    for key, val in self.comm.bcast(None, root=0):
                        self._jacobian[key] = val
            else:
                if self._discrete_inputs:
                    self.compute_partials(self._inputs, self._jacobian, self._discrete_inputs)
                else:
                    self.compute_partials(self._inputs, self._jacobian)

    def _linearize(self, jac=None, sub_do_ln=False):
        """
        Compute jacobian / factorization. The model is assumed to be in a scaled state.

        Parameters
        ----------
        jac : Jacobian or None
            Ignored.
        sub_do_ln : bool
            Flag indicating if the children should call linearize on their linear solvers.
        """
        if self.matrix_free or not (self._has_compute_partials or self._approx_schemes):
            return

        self._check_first_linearize()

        with self._unscaled_context(outputs=[self._outputs], residuals=[self._residuals]):
            # Computing the approximation before the call to compute_partials allows users to
            # override FD'd values.
            for approximation in self._approx_schemes.values():
                approximation.compute_approximations(self, jac=self._jacobian)

            if self._has_compute_partials:
                # We used to negate the jacobian here, and then re-negate after the hook.
                self._compute_partials_wrapper()

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute outputs given inputs. The model is assumed to be in an unscaled state.

        An inherited component may choose to either override this function or to define a
        compute_primal function.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : dict-like or None
            If not None, dict-like object containing discrete input values.
        discrete_outputs : dict-like or None
            If not None, dict-like object containing discrete output values.
        """
        global _tuplist

        if self.compute_primal is None:
            return

        returns = \
            self.compute_primal(*self._get_compute_primal_invals(inputs, discrete_inputs))

        if not isinstance(returns, _tuplist):
            returns = (returns,)

        if not discrete_outputs:
            outputs.set_vals(returns)
        else:
            outputs.set_vals(returns[:outputs.nvars()])
            self._discrete_outputs.set_vals(returns[outputs.nvars():])

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Sub-jac components written to partials[output_name, input_name]..
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        """
        pass

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode, discrete_inputs=None):
        r"""
        Compute jac-vector product. The model is assumed to be in an unscaled state.

        If mode is:
            'fwd': d_inputs \|-> d_outputs

            'rev': d_outputs \|-> d_inputs

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        d_inputs : Vector
            See inputs; product must be computed only if var_name in d_inputs.
        d_outputs : Vector
            See outputs; product must be computed only if var_name in d_outputs.
        mode : str
            Either 'fwd' or 'rev'.
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        """
        pass

    def is_explicit(self):
        """
        Return True if this is an explicit component.

        Returns
        -------
        bool
            True if this is an explicit component.
        """
        return True

    def _get_compute_primal_invals(self, inputs, discrete_inputs):
        """
        Yield the inputs expected by the compute_primal method.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables Vector.
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.

        Yields
        ------
        any
            Inputs expected by the compute_primal method.
        """
        yield from inputs.values()
        if discrete_inputs:
            yield from discrete_inputs.values()

    def _get_compute_primal_argnames(self):
        """
        Return the expected argnames for the compute_primal method.

        Returns
        -------
        list
            List of argnames expected by the compute_primal method.
        """
        argnames = []
        argnames.extend(self._var_rel_names['input'])
        if self._discrete_inputs:
            argnames.extend(self._discrete_inputs)
        return argnames

    def _setup_jax(self, from_group=False):
        """
        Set up the jax interface for this component.

        Parameters
        ----------
        from_group : bool
            If True, this is being called from a Group setup.
        """
        if self.matrix_free is True:
            self.compute_jacvec_product = MethodType(_jax_compute_jacvec_product, self)
        else:
            self.compute_partials = MethodType(_jax_compute_partials, self)
            self._has_compute_partials = True

        if self.compute_primal is None:
            # convert the compute method to a compute_primal method
            jaxifier = ExplicitCompJaxify(self, verbose=True)

            if jaxifier.get_self_statics:
                self.get_self_statics = MethodType(jaxifier.get_self_statics, self)
            # replace existing compute method with base class method, so that compute_primal
            # will be called.
            self.compute = MethodType(ExplicitComponent.compute, self)

            self.compute_primal = MethodType(jaxifier.compute_primal, self)
            self._compute_primal_returns_tuple = True
        else:
            # check that compute_primal args are in the correct order
            args = list(inspect.signature(self.compute_primal).parameters)
            if args and args[0] == 'self':
                args = args[1:]
            compargs = self._get_compute_primal_argnames()
            if args != compargs:
                raise RuntimeError(f"{self.msginfo}: compute_primal method args {args} don't match "
                                   f"the expected args {compargs}.")

            # determine if the compute_primal method returns a tuple
            self._compute_primal_returns_tuple = ReturnChecker(self.compute_primal).returns_tuple()

        if not from_group and self.options['use_jit']:
            static_argnums = []
            idx = len(self._var_rel_names['input']) + 1
            static_argnums.extend(range(idx, idx + len(self._discrete_inputs)))
            self.compute_primal = MethodType(jit(self.compute_primal.__func__,
                                                 static_argnums=static_argnums), self)

        _jax_register_pytree_class(self.__class__)

    def _get_jac_func(self):
        """
        Return the jacobian function for this component.

        In forward mode, jax.jacfwd is used, and in reverse mode, jax.jacrev is used.  The direction
        is chosen automatically based on the sizes of the inputs and outputs.

        Returns
        -------
        function
            The jacobian function.
        """
        # TODO: modify this to use relevance and possibly compile multiple jac functions depending
        # on DV/response so that we don't compute any derivatives that are always zero.
        if self._jac_func_ is None:
            fjax = jax.jacfwd if self.best_partial_deriv_direction() == 'fwd' else jax.jacrev
            nstatic = len(self._discrete_inputs)
            wrt_idxs = list(range(1, len(self._var_abs2meta['input']) + 1))
            self._jac_func_ = MethodType(fjax(self.compute_primal.__func__, argnums=wrt_idxs), self)

            if self.options['use_jit']:
                static_argnums = tuple(range(1 + len(wrt_idxs), 1 + len(wrt_idxs) + nstatic))
                self._jac_func_ = MethodType(jit(self._jac_func_.__func__,
                                                 static_argnums=static_argnums),
                                             self)

        return self._jac_func_



"""Test serial derivatives in implicit group when running in parallel"""
import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.mpi import MPI  # MPI will be None here if MPI is not active

try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class SerialImplicitDerivConsist(unittest.TestCase):
    N_PROCS = 2

    def test_serialImplicitDerivConsist(self):
        class Imp(om.ImplicitComponent):
            def setup(self):
                self.add_input("x", shape_by_conn=True, distributed=True)
                self.add_input("s0")
                self.add_input("s1")
                self.add_output(
                    "y", shape=5 if self.comm.rank == 0 else 0, distributed=True
                )

            def apply_nonlinear(self, inputs, outputs, residuals):
                residuals["y"] = outputs["y"] - (
                    np.sum(inputs["x"]) + inputs["s0"] + inputs["s1"]
                )

            def solve_nonlinear(self, inputs, outputs):
                outputs["y"] = np.sum(inputs["x"]) + inputs["s0"] + inputs["s1"]

            def apply_linear(
                self, inputs, outputs, d_inputs, d_outputs, d_residuals, mode
            ):
                if mode == "rev":
                    if "y" in d_residuals:
                        if "x" in d_inputs:
                            d_inputs["x"] = -np.sum(d_residuals["y"])
                        if "s0" in d_inputs:
                            d_inputs["s0"] = self.comm.allreduce(
                                -np.sum(d_residuals["y"])
                            )
                        if "s1" in d_inputs:
                            d_inputs["s1"] = self.comm.allreduce(
                                -np.sum(d_residuals["y"])
                            )
                        if "y" in d_outputs:
                            d_outputs["y"] = d_residuals["y"]

        class Exp(om.ExplicitComponent):
            def setup(self):
                self.add_input("y", distributed=True, shape_by_conn=True)
                self.add_output("total")

            def compute(self, inputs, outputs):
                outputs["total"] = self.comm.allreduce(np.sum(inputs["y"]))

            def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
                if mode == "rev":
                    d_inputs["y"] += d_outputs["total"]

        prob = om.Problem()
        ivc = prob.model.add_subsystem("ivc", om.IndepVarComp(), promotes=["*"])
        ivc.add_output("x", val=1, distributed=True)
        ivc.add_output("s0", val=2.0)
        ivc.add_output("s1", val=3.0)

        prob.model.add_subsystem("Imp", Imp(), promotes=["*"])
        prob.model.add_subsystem("Exp", Exp(), promotes=["*"])

        prob.setup(mode="rev")
        prob.run_model()
        totals = prob.check_totals(of="total", wrt="s0")
        for var, err in totals.items():
            assert_near_equal(err["rel error"].reverse, 0.0, 5e-3)


if __name__ == "__main__":
    unittest.main()



"""
collection of complex-step safe functions to replace standard numpy operations.
"""

import numpy as np

if np.__version__[0] == '2':
    NumPy2 = True
else:
    NumPy2 = False


def abs(x):
    """
    Complex-step safe version of numpy.abs function.

    Parameters
    ----------
    x : ndarray
        Array value to be computed on.

    Returns
    -------
    ndarray
        Absolute value.
    """
    if isinstance(x, np.ndarray):
        # Changed in NumPy 2.0: Definition of complex sign changed to follow the Array API standard.
        # 1.x: For complex inputs, the sign function returns sign(x.real) + 0j if x.real != 0
        #      else sign(x.imag) + 0j.
        # 2.0: For complex inputs, the sign function returns x / abs(x), and 0 if x==0.
        if NumPy2 and np.any(np.iscomplex(x)):
            # replicate NumPy 1.x behavior for complex arrays
            z0_idx = x.real == 0
            nz_idx = x.real != 0
            signs = np.zeros(x.shape, dtype=complex)
            signs[nz_idx] = np.sign(x[nz_idx]).real + 0j
            signs[z0_idx] = np.sign(x[z0_idx].imag) + 0j
            return x * signs
        else:
            return x * np.sign(x)
    elif x.real < 0.0:
        return -x
    return x


def norm(x, axis=None):
    """
    Complex-step safe version of numpy.linalg.norm function.

    Parameters
    ----------
    x : ndarray
        Array value to be computed on.
    axis : None or int
        Axis to perform the norm up to.

    Returns
    -------
    ndarray
        Matrix or vector norm.
    """
    return np.sqrt(np.sum(x**2, axis=axis))


def arctan2(y, x):
    """
    Numpy-compatible, complex-compatible arctan2 function for use with complex-step.

    Parameters
    ----------
    y : float or complex
        The length of the side opposite the angle being determined.
    x : float or complex
        The length of the side adjacent to the angle being determined.

    Returns
    -------
    ndarray
        The angle whose opposite side has length y and whose adjacent side has length x.
    """
    if np.iscomplexobj(x) or np.iscomplexobj(y):
        a = np.real(y)
        b = np.imag(y)
        c = np.real(x)
        d = np.imag(x)
        return np.arctan2(a, c) + 1j * (c * b - a * d) / (a**2 + c**2)
    else:
        return np.arctan2(y, x)



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

from openmdao.test_suite.components.paraboloid import Paraboloid

from openmdao.visualization.case_viewer.case_viewer import _apply_slice, _apply_transform, \
    _get_var_meta, _get_opt_vars, _get_vars, _get_resids_vars, _get_resids_val, _get_plot_style


@use_tempdirs
class TestCaseRetrieval(unittest.TestCase):

    def setUp(self):
        # build the model
        prob = self._prob = om.Problem()
        prob.model.add_subsystem('parab', Paraboloid(), promotes_inputs=['x', 'y'])

        # define the component whose output will be constrained
        prob.model.add_subsystem('const', om.ExecComp('g = x + y'), promotes_inputs=['x', 'y'])

        # Design variables 'x' and 'y' span components, so we need to provide a common initial
        # value for them.
        prob.model.set_input_defaults('x', 3.0)
        prob.model.set_input_defaults('y', -4.0)

        # setup the optimization
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'

        recorder = om.SqliteRecorder('parab_record.sql')
        prob.driver.add_recorder(recorder)
        prob.model.add_recorder(recorder)
        prob.add_recorder(recorder)

        prob.driver.recording_options['includes'] = ['*']
        prob.driver.recording_options['record_desvars'] = True
        prob.driver.recording_options['record_objectives'] = True
        prob.driver.recording_options['record_constraints'] = True
        prob.driver.recording_options['record_residuals'] = True

        prob.model.add_design_var('x', lower=-50, upper=50)
        prob.model.add_design_var('y', lower=-50, upper=50)
        prob.model.add_objective('parab.f_xy')

        # to add the constraint to the model
        prob.model.add_constraint('const.g', lower=0, upper=10.)

        prob.setup()

        prob.run_driver()

        prob.record("final")

    def test_get_vars(self):
        cr = om.CaseReader(self._prob.get_outputs_dir() / 'parab_record.sql')
        case_names = cr.list_cases(out_stream=None)
        inputs = _get_vars(cr, case_names=case_names, var_types='inputs')
        outputs = _get_vars(cr, case_names=case_names, var_types='outputs')
        self.assertSetEqual(set(inputs), {'x', 'y'})
        self.assertSetEqual(set(outputs), {'const.g', 'parab.f_xy'})

    def test_get_opt_vars(self):
        cr = om.CaseReader(self._prob.get_outputs_dir() / 'parab_record.sql')
        case_names = cr.list_cases(out_stream=None)
        desvars = _get_opt_vars(cr, case_names, var_type='desvars')
        constraints = _get_opt_vars(cr, case_names, var_type='constraints')
        objectives = _get_opt_vars(cr, case_names, var_type='objectives')

        self.assertSetEqual(set(desvars), {'x', 'y'})
        self.assertSetEqual(set(constraints), {'const.g'})
        self.assertSetEqual(set(objectives), {'parab.f_xy'})

        all = _get_opt_vars(cr, case_names)
        self.assertSetEqual(set(all), {'x', 'y', 'const.g', 'parab.f_xy'})

    def test_get_var_meta(self):
        cr = om.CaseReader(self._prob.get_outputs_dir() / 'parab_record.sql')
        case_name = cr.list_cases(out_stream=None)[0]
        g_meta = _get_var_meta(cr, case_name, 'const.g')
        self.assertEqual(g_meta['prom_name'], 'const.g')
        self.assertEqual(g_meta['units'], None)
        self.assertEqual(g_meta['shape'], (1,))

        f_xy_meta = _get_var_meta(cr, case_name, 'parab.f_xy')
        self.assertEqual(f_xy_meta['prom_name'], 'parab.f_xy')
        self.assertEqual(f_xy_meta['units'], None)
        self.assertEqual(f_xy_meta['shape'], (1,))

        x_meta = _get_var_meta(cr, case_name, 'x')
        self.assertEqual(x_meta['prom_name'], 'x')
        self.assertEqual(x_meta['units'], None)
        self.assertEqual(x_meta['shape'], (1,))

    def test_get_resids_vars(self):
        cr = om.CaseReader(self._prob.get_outputs_dir() / 'parab_record.sql')
        case_names = cr.list_cases(out_stream=None)
        vars_with_resids = _get_resids_vars(cr, case_names)
        self.assertSetEqual(set(vars_with_resids), {'const.g', 'parab.f_xy'})

    def test_get_resids_val(self):
        cr = om.CaseReader(self._prob.get_outputs_dir() / 'parab_record.sql')
        case_name = cr.list_cases(out_stream=None)[0]
        case = cr.get_case(case_name)

        g_resids = _get_resids_val(case, 'const.g')
        f_resids = _get_resids_val(case, 'parab.f_xy')

        assert_near_equal(g_resids, [0.])
        assert_near_equal(f_resids, [0.])


class TestOtherFuncs(unittest.TestCase):

    def test_apply_slice(self):
        x = np.random.random((20, 3)) - 0.5

        assert_near_equal(_apply_slice(x, '[...]'), x[...])
        assert_near_equal(_apply_slice(x, '[2, 1::2]'), x[2, 1::2])
        assert_near_equal(_apply_slice(x, '[:, -1]'), x[:, -1])
        assert_near_equal(_apply_slice(x, '[:, [0, 1]]'), x[:, [0, 1]])

    def test_apply_transform(self):
        x = np.random.random((20, 3)) - 0.5

        max = np.max(x)
        assert_near_equal(_apply_transform(x, 'max'), max)

        min = np.min(x)
        assert_near_equal(_apply_transform(x, 'min'), min)

        maxabs = np.max(np.abs(x))
        assert_near_equal(_apply_transform(x, 'maxabs'), maxabs)

        minabs = np.min(np.abs(x))
        assert_near_equal(_apply_transform(x, 'minabs'), minabs)

        norm = np.linalg.norm(x)
        assert_near_equal(_apply_transform(x, 'norm'), norm)

        ravel = np.ravel(x)
        assert_near_equal(_apply_transform(x, 'ravel'), ravel)

        assert_near_equal(_apply_transform(x, 'None'), x)

    def test_get_plot_style(self):
        # Selected case index is greater or equal to than the number of cases.
        # Style all plots equally.
        lw, ms, s, alpha = _get_plot_style(1, 20, 20)
        self.assertEqual(lw, 1.0)
        self.assertEqual(ms, 2.0)
        self.assertEqual(s, 10)
        self.assertEqual(alpha, 1.0)

        # Selected case index is less than the number of cases
        # but idx != selected_case idx
        # This plot should have low visibility.
        lw, ms, s, alpha = _get_plot_style(1, 19, 20)
        self.assertEqual(lw, 0.5)
        self.assertEqual(ms, 2.0)
        self.assertEqual(s, 10)
        self.assertEqual(alpha, 0.1)

        # Selected case index is less than the number of cases
        # and idx == selected_case idx
        # This plot should have high visibility.
        lw, ms, s, alpha = _get_plot_style(19, 19, 20)
        self.assertEqual(lw, 2.0)
        self.assertEqual(ms, 6.0)
        self.assertEqual(s, 40)
        self.assertEqual(alpha, 1)



import unittest
from tempfile import TemporaryFile

import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.sellar import SellarDis1, SellarDis2, SellarDerivativesGrouped
from openmdao.error_checking.check_config import get_sccs_topo, _all_non_redundant_checks
from openmdao.utils.assert_utils import assert_warning, assert_no_warning
from openmdao.utils.logger_utils import TestLogger
from openmdao.utils.testing_utils import use_tempdirs


class MyComp(om.ExecComp):
    def __init__(self):
        super().__init__(["y = 2.0*a", "z = 3.0*b"])


class TestCheckConfig(unittest.TestCase):

    def test_log_messages(self):
        p = om.Problem(SellarDerivativesGrouped())

        testlogger = TestLogger()
        p.setup(check='all', logger=testlogger)
        p.final_setup()

        for check in sorted(_all_non_redundant_checks):
            testlogger.find_in('info', f'checking {check}...')
            testlogger.find_match_in('info', f'    {check} check complete (* sec).')

    def test_dataflow_1_level(self):
        p = om.Problem()
        root = p.model

        root.add_subsystem("indep", om.IndepVarComp('x', 1.0))
        root.add_subsystem("C1", MyComp())
        root.add_subsystem("C2", MyComp())
        root.add_subsystem("C3", MyComp())
        root.add_subsystem("C4", MyComp())

        root.connect("C4.y", "C2.a")
        root.connect("C4.y", "C3.a")
        root.connect("C2.y", "C1.a")
        root.connect("C1.y", "C4.a")

        # make sure no system has dangling inputs so we avoid that warning
        root.connect("indep.x", "C1.b")
        root.connect("indep.x", "C2.b")
        root.connect("indep.x", "C3.b")
        root.connect("indep.x", "C4.b")

        # set iterative solvers since we have cycles
        root.linear_solver = om.LinearBlockGS()
        root.nonlinear_solver = om.NonlinearBlockGS()

        testlogger = TestLogger()
        p.setup(check=['cycles', 'out_of_order'], logger=testlogger)
        p.final_setup()

        testlogger.find_in('info', "The following groups contain cycles:")
        testlogger.find_in('info', "   Group '' has the following cycles:")
        testlogger.find_in('info', "      ['C1', 'C2', 'C4']")
        testlogger.find_in('warning', "The following systems are executed out-of-order:\n"
            "   System 'C3' executes out-of-order with respect to its source systems ['C4']\n")

    def test_parallel_group_order(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 1.0))
        model.add_subsystem('p2', om.IndepVarComp('x', 1.0))

        parallel = model.add_subsystem('parallel', om.ParallelGroup())
        parallel.add_subsystem('c2', om.ExecComp(['y=5.0*x']))
        parallel.add_subsystem('c1', om.ExecComp(['y=-2.0*x']))
        parallel.connect('c1.y', 'c2.x')

        parallel = model.add_subsystem('parallel_copy', om.ParallelGroup())
        parallel.add_subsystem('comp2', om.ExecComp(['y=5.0*x']))
        parallel.add_subsystem('comp1', om.ExecComp(['y=-2.0*x']))
        parallel.connect('comp1.y', 'comp2.x')

        model.add_subsystem('c3', om.ExecComp(['y=3.0*x1+7.0*x2']))
        model.add_subsystem('c4', om.ExecComp(['y=3.0*x_copy_1+7.0*x_copy_2']))

        model.connect("parallel.c1.y", "c3.x1")
        model.connect("parallel.c2.y", "c3.x2")
        model.connect("parallel_copy.comp1.y", "c4.x_copy_1")
        model.connect("parallel_copy.comp2.y", "c4.x_copy_2")

        model.connect("p1.x", "parallel.c1.x")
        model.connect("p1.x", "parallel_copy.comp1.x")

        testlogger = TestLogger()
        prob.setup(check=True, mode='fwd', logger=testlogger)

        msg = "Need to attach NonlinearBlockJac, NewtonSolver, or BroydenSolver to 'parallel' when " \
              "connecting components inside parallel groups"

        with assert_warning(UserWarning, msg):
            prob.run_model()

        for w in testlogger.get('warning'):
            self.assertTrue('out-of-order' not in w)

    def test_serial_in_parallel(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 1.0))

        parallel = model.add_subsystem('parallel', om.ParallelGroup())
        parallel.add_subsystem('c1', om.ExecComp(['y=-2.0*x']))

        parallel2 = model.add_subsystem('parallel_copy', om.ParallelGroup())
        parallel2.add_subsystem('comp1', om.ExecComp(['y=-2.0*x']))

        model.add_subsystem('con', om.ExecComp('y = 3.0*x'))

        model.connect("p1.x", "parallel.c1.x")
        model.connect('parallel.c1.y', 'parallel_copy.comp1.x')
        model.connect('parallel_copy.comp1.y', 'con.x')

        prob.setup(check=True)

        msg = ("The following systems are executed out-of-order:\n"
               "   System 'parallel.c2' executes out-of-order with respect to its source systems ['parallel.c1']\n")

        with assert_no_warning(UserWarning, msg):
            prob.run_model()

    def test_single_parallel_group_order(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 1.0))
        model.add_subsystem('p2', om.IndepVarComp('x', 1.0))

        parallel = model.add_subsystem('parallel', om.ParallelGroup())
        parallel.add_subsystem('c2', om.ExecComp(['y=5.0*x']))
        parallel.add_subsystem('c1', om.ExecComp(['y=-2.0*x']))
        parallel.connect('c1.y', 'c2.x')

        model.add_subsystem('c3', om.ExecComp(['y=3.0*x1+7.0*x2']))

        model.connect("parallel.c1.y", "c3.x1")
        model.connect("parallel.c2.y", "c3.x2")

        model.connect("p1.x", "parallel.c1.x")

        testlogger = TestLogger()
        prob.setup(check=True, mode='fwd', logger=testlogger)

        msg = "Need to attach NonlinearBlockJac, NewtonSolver, or BroydenSolver to 'parallel' when " \
              "connecting components inside parallel groups"

        with assert_warning(UserWarning, msg):
            prob.run_model()

        for w in testlogger.get('warning'):
            self.assertTrue('out-of-order' not in w)

    def test_no_connect_parallel_group(self):
        prob = om.Problem()
        model = prob.model

        traj = model.add_subsystem('traj', om.ParallelGroup())

        burn1 = traj.add_subsystem('burn1', om.Group())
        burn1.add_subsystem('p1', om.IndepVarComp('x', 1.0))
        burn1.add_subsystem('burn_eq1', om.ExecComp(['y=-2.0*x']))
        burn1.connect('p1.x', 'burn_eq1.x')

        burn2 = traj.add_subsystem('burn2', om.Group())
        burn2.add_subsystem('p2', om.IndepVarComp('x', 1.0))
        burn2.add_subsystem('burn_eq2', om.ExecComp(['y=5.0*x']))
        burn2.connect('p2.x', 'burn_eq2.x')

        testlogger = TestLogger()
        prob.setup(check=True, mode='fwd', logger=testlogger)

        msg = "Need to attach NonlinearBlockJac, NewtonSolver, or BroydenSolver to 'parallel' when " \
              "connecting components inside parallel groups"

        with assert_no_warning(UserWarning, msg):
            prob.run_model()

    def test_dataflow_multi_level(self):
        p = om.Problem()
        root = p.model

        root.add_subsystem("indep", om.IndepVarComp('x', 1.0))

        G1 = root.add_subsystem("G1", om.Group())

        G1.add_subsystem("C1", MyComp())
        G1.add_subsystem("C2", MyComp())

        root.add_subsystem("C3", MyComp())
        root.add_subsystem("C4", MyComp())

        root.connect("C4.y", "G1.C2.a")
        root.connect("C4.y", "C3.a")
        root.connect("G1.C2.y", "G1.C1.a")
        root.connect("G1.C1.y", "C4.a")

        # make sure no system has dangling inputs so we avoid that warning
        root.connect("indep.x", "G1.C1.b")
        root.connect("indep.x", "G1.C2.b")
        root.connect("indep.x", "C3.b")
        root.connect("indep.x", "C4.b")

        # set iterative solvers since we have cycles
        root.linear_solver = om.LinearBlockGS()
        root.nonlinear_solver = om.NonlinearBlockGS()

        testlogger = TestLogger()
        p.setup(check=['cycles', 'out_of_order'], logger=testlogger)
        p.final_setup()

        expected_info = (
            "The following groups contain cycles:",
            "   Group '' has the following cycles:",
            "      ['G1', 'C4']"
        )

        expected_warning = (
            "The following systems are executed out-of-order:\n"
            "   In System 'G1', subsystem 'C1' executes out-of-order with respect to its source systems ['C2']\n"
            "   System 'C3' executes out-of-order with respect to its source systems ['C4']\n"
        )

        for msg in expected_info:
            testlogger.find_in('info', msg)
        testlogger.find_in('warning', expected_warning)

        # test comps_only cycle check
        graph = root.compute_sys_graph(comps_only=True)
        sccs = [sorted(s) for s in get_sccs_topo(graph) if len(s) > 1]
        self.assertEqual([['C4', 'G1.C1', 'G1.C2']], sccs)

    def test_out_of_order_repeat_bug_and_dup_inputs(self):
        p = om.Problem()
        p.model.add_subsystem("indep", om.IndepVarComp('x', 1.0))
        p.model.add_subsystem("C1", om.ExecComp(["y = 2.0*a", "z = 3.0*b"]))
        p.model.add_subsystem("C2", om.ExecComp("y = 2.0*a"))

        # create 2 out of order connections from C2 to C1
        p.model.connect("C2.y", "C1.a")
        p.model.connect("C2.y", "C1.b")

        # make sure no system has dangling inputs so we avoid that warning
        p.model.connect("indep.x", "C2.a")

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        expected_warning_1 = (
            "The following systems are executed out-of-order:\n"
            "   System 'C1' executes out-of-order with respect to its source systems ['C2']\n"
        )

        expected_warning_2 = (
            "The following components have multiple inputs connected to the same source, "
            "which can introduce unnecessary data transfer overhead:\n"
            "   C1 has inputs ['a', 'b'] connected to C2.y\n"
        )

        testlogger.find_in('warning', expected_warning_1)
        testlogger.find_in('warning', expected_warning_2)

    def test_multi_cycles(self):
        p = om.Problem()
        root = p.model

        root.add_subsystem("indep", om.IndepVarComp('x', 1.0))

        def make_cycle(root, start, end):
            # systems within a cycle will be declared out of order, but
            # should not be reported since they're internal to a cycle.
            for i in range(end, start-1, -1):
                root.add_subsystem("C%d" % i, MyComp())

            for i in range(start, end):
                root.connect("C%d.y" % i, "C%d.a" % (i+1))
            root.connect("C%d.y" % end, "C%d.a" % start)

        G1 = root.add_subsystem('G1', om.Group())

        make_cycle(G1, 1, 3)

        G1.add_subsystem("N1", MyComp())

        make_cycle(G1, 11, 13)

        G1.add_subsystem("N2", MyComp())

        make_cycle(G1, 21, 23)

        G1.add_subsystem("N3", MyComp())

        G1.connect("N1.z", "C12.b")
        G1.connect("C13.z", "N2.b")
        G1.connect("N2.z", "C21.b")
        G1.connect("C23.z", "N3.b")
        G1.connect("N3.z", "C2.b")
        G1.connect("C11.z", "C3.b")

        # set iterative solvers since we have cycles
        root.linear_solver = om.LinearBlockGS()
        root.nonlinear_solver = om.NonlinearBlockGS()

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        expected_warning_1 = (
            "The following systems are executed out-of-order:\n"
            "   In System 'G1', subsystem 'C2' executes out-of-order with respect to its source systems ['N3']\n"
            "   In System 'G1', subsystem 'C3' executes out-of-order with respect to its source systems ['C11']\n"
        )

        testlogger.find_in('warning', expected_warning_1)

    def test_multi_cycles_non_default(self):
        p = om.Problem()
        root = p.model

        root.add_subsystem("indep", om.IndepVarComp('x', 1.0))

        def make_cycle(root, start, end):
            # systems within a cycle will be declared out of order, but
            # should not be reported since they're internal to a cycle.
            for i in range(end, start-1, -1):
                root.add_subsystem("C%d" % i, MyComp())

            for i in range(start, end):
                root.connect("C%d.y" % i, "C%d.a" % (i+1))
            root.connect("C%d.y" % end, "C%d.a" % start)

        G1 = root.add_subsystem('G1', om.Group())

        make_cycle(G1, 1, 3)

        G1.add_subsystem("N1", MyComp())

        make_cycle(G1, 11, 13)

        G1.add_subsystem("N2", MyComp())

        make_cycle(G1, 21, 23)

        G1.add_subsystem("N3", MyComp())

        G1.connect("N1.z", "C12.b")
        G1.connect("C13.z", "N2.b")
        G1.connect("N2.z", "C21.b")
        G1.connect("C23.z", "N3.b")
        G1.connect("N3.z", "C2.b")
        G1.connect("C11.z", "C3.b")

        # set iterative solvers since we have cycles
        root.linear_solver = om.LinearBlockGS()
        root.nonlinear_solver = om.NonlinearBlockGS()

        testlogger = TestLogger()
        p.setup(check=['cycles', 'out_of_order', 'unconnected_inputs'], logger=testlogger)
        p.final_setup()

        expected_info = (
            "The following groups contain cycles:",
            "   Group 'G1' has the following cycles:",
            "      ['C13', 'C12', 'C11']",
            "      ['C23', 'C22', 'C21']",
            "      ['C3', 'C2', 'C1']"
        )

        expected_warning_1 = (
            "The following systems are executed out-of-order:\n"
            "   In System 'G1', subsystem 'C2' executes out-of-order with respect to its source systems ['N3']\n"
            "   In System 'G1', subsystem 'C3' executes out-of-order with respect to its source systems ['C11']\n"
        )

        for msg in expected_info:
            testlogger.find_in('info', msg)
        testlogger.find_in('warning', expected_warning_1)

    def test_unconnected_auto_ivc(self):
        class SellarMDAConnect(om.Group):

            def setup(self):
                cycle = self.add_subsystem('cycle', om.Group(), promotes_inputs=['x', 'z'])
                cycle.add_subsystem('d1', SellarDis1(), promotes_inputs=['x', 'z'])
                cycle.add_subsystem('d2', SellarDis2(), promotes_inputs=['z'])
                cycle.connect('d1.y1', 'd2.y1')

                cycle.nonlinear_solver = om.NonlinearBlockGS()

                self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                          z=np.array([0.0, 0.0]), x=0.0),
                                   promotes_inputs=['x', 'z'])

                self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'))
                self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'))

                self.connect('cycle.d1.y1', ['obj_cmp.y1', 'con_cmp1.y1'])
                self.connect('cycle.d2.y2', ['obj_cmp.y2', 'con_cmp2.y2'])

        prob = om.Problem(model=SellarMDAConnect())

        prob.model.add_design_var('z', lower=0, upper=10)
        prob.model.add_objective('obj_cmp.obj')
        prob.model.add_constraint('con_cmp1.con1', upper=0)
        prob.model.add_constraint('con_cmp2.con2', upper=0)

        testlogger = TestLogger()
        prob.setup(check=['unconnected_inputs'], logger=testlogger)
        prob.final_setup()

        expected_warning = (
            "The following inputs are connected to Auto IVC output variables:\n"
            "  cycle.d1.y2 (cycle.d1.y2)\n"
            "  x (cycle.d1.x)\n"
            "  x (obj_cmp.x)\n"
        )

        testlogger.find_in('warning', expected_warning)

    def test_comp_has_no_outputs(self):
        p = om.Problem()
        root = p.model

        root.add_subsystem("indep", om.IndepVarComp('x', 1.0))

        comp1 = root.add_subsystem("comp1", om.ExplicitComponent())
        comp1.add_input('x', val=0.)

        root.connect('indep.x', 'comp1.x')

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        expected = (
            "The following Components do not have any outputs:\n"
            "   comp1\n"
        )

        testlogger.find_in('warning', expected)

    def test_initial_condition_order(self):
        # Makes sure we set vars to their initial condition before running checks.

        class TestComp(om.ExplicitComponent):

            def setup(self):
                self.add_input('x', 37.0)
                self.add_output('y', 45.0)

            def check_config(self, logger):
                x = self._vectors['input']['nonlinear']['x']
                if x != 75.0:
                    raise ValueError('Check config is being called before initial conditions are set.')

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                outputs['y'] = inputs['x']

        prob = om.Problem()
        prob.model.add_subsystem('comp', TestComp())

        prob.setup(check='all')

        prob['comp.x'] = 75.0

        prob.final_setup()

    def test_unserializable_option(self):
        # Makes sure we get a warning if an option is not picklable.

        class TestComp(om.ExplicitComponent):
            def initialize(self):
                self.options.declare('file1')
                self.options.declare('file2', recordable=False)

            def setup(self):
                self.options['file1'] = TemporaryFile()
                self.options['file2'] = TemporaryFile()

                self.add_input('x', 0.)
                self.add_output('y', 0.)

            def compute(self, inputs, outputs):
                outputs['y'] = inputs['x']

        class TestNewton(om.NewtonSolver):
            def _declare_options(self):
                super()._declare_options()
                self.options.declare('file3')

        class TestLinesearch(om.BoundsEnforceLS):
            def _declare_options(self):
                super()._declare_options()
                self.options.declare('file4', recordable=False)

        class TestKrylov(om.ScipyKrylov):
            def _declare_options(self):
                super()._declare_options()
                self.options.declare('file5')

        prob = om.Problem()
        prob.model.add_subsystem('comp', TestComp())
        prob.model.nonlinear_solver = TestNewton(solve_subsystems=False, file3=TemporaryFile())
        prob.model.nonlinear_solver.linesearch = TestLinesearch(file4=TemporaryFile())
        prob.model.nonlinear_solver.linear_solver = TestKrylov(file5=TemporaryFile())

        msg1 = "'comp' <class TestComp>: option 'file1' is not serializable " \
               "(cannot be pickled) but 'recordable=False' has not been set. " \
               "No options will be recorded for this TestComp unless " \
               "'recordable' is set to False for this option."

        msg2 = "'comp' <class TestComp>: option 'file2' is not serializable " \
               "(cannot be pickled) and will not be recorded."

        msg3 = "TestNewton in <model> <class Group>: nonlinear_solver option " \
               "'file3' is not serializable (cannot be pickled) but 'recordable=False' " \
               "has not been set. No options will be recorded for this TestNewton unless " \
               "'recordable' is set to False for this option."

        msg4 = "TestLinesearch in <model> <class Group>: linesearch option 'file4' is " \
               "not serializable (cannot be pickled) and will not be recorded."

        msg5 = "TestKrylov in <model> <class Group>: linear_solver option 'file5' " \
               "is not serializable (cannot be pickled) but 'recordable=False' has " \
               "not been set. No options will be recorded for this TestKrylov unless " \
               "'recordable' is set to False for this option."

        # By default, a warning is only issued if 'recordable' is not set to False
        testlogger = TestLogger()
        prob.setup(check=True, logger=testlogger)
        prob.final_setup()

        testlogger.find_in('warning', msg1)
        testlogger.find_in('warning', msg3)
        testlogger.find_in('warning', msg5)

        self.assertFalse(testlogger.contains('warning', msg2))
        self.assertFalse(testlogger.contains('warning', msg4))

        # When checking 'all_unserializable_options', a warning is issued even if recordable=False
        testlogger = TestLogger()
        prob.setup(check=['all_unserializable_options'], logger=testlogger)
        prob.final_setup()

        testlogger.find_in('warning', msg1)
        testlogger.find_in('warning', msg2)
        testlogger.find_in('warning', msg3)
        testlogger.find_in('warning', msg4)
        testlogger.find_in('warning', msg5)


@use_tempdirs
class TestRecorderCheckConfig(unittest.TestCase):

    def setUp(self):
        self.recorder = om.SqliteRecorder("sqlite_test")

    def test_check_no_recorder_set(self):
        p = om.Problem()

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        expected_warning = "The Problem has no recorder of any kind attached"
        testlogger.find_in('warning', expected_warning)

    def test_check_problem_recorder_set(self):
        p = om.Problem()
        p.add_recorder(self.recorder)

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        warnings = testlogger.get('warning')
        self.assertEqual(len(warnings), 0)

    def test_check_driver_recorder_set(self):
        p = om.Problem()
        p.driver.add_recorder(self.recorder)

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        warnings = testlogger.get('warning')
        self.assertEqual(len(warnings), 0)

    def test_check_system_recorder_set(self):
        p = om.Problem()
        p.model.add_recorder(self.recorder)

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()

        warnings = testlogger.get('warning')
        self.assertEqual(len(warnings), 0)

    def test_check_linear_solver_recorder_set(self):
        p = om.Problem()
        p.model.nonlinear_solver.add_recorder(self.recorder)

        testlogger = TestLogger()
        p.setup(check=True, logger=testlogger)
        p.final_setup()
        p.cleanup()

        warnings = testlogger.get('warning')
        self.assertEqual(len(warnings), 0)


if __name__ == "__main__":
    unittest.main()



"""Various debugging functions."""


import sys
from io import StringIO

import numpy as np
from contextlib import contextmanager
from collections import Counter

from openmdao.core.constants import _SetupStatus, _DEFAULT_OUT_STREAM
from openmdao.utils.mpi import MPI
from openmdao.utils.reports_system import register_report
from openmdao.utils.file_utils import text2html, _load_and_exec
from openmdao.utils.rangemapper import RangeMapper
from openmdao.visualization.tables.table_builder import generate_table


class _NoColor(object):
    """
    A class to replace Fore, Back, and Style when colorama isn't istalled.
    """
    def __getattr__(self, name):
        return ''


def _get_color_printer(stream=sys.stdout, colors=True, rank=0):
    """
    Return a print function tied to a particular stream, along with coloring info.
    """
    try:
        from colorama import init, Fore, Back, Style
        init(autoreset=True)
    except ImportError:
        Fore = Back = Style = _NoColor()

    if not colors:
        Fore = Back = Style = _NoColor()

    if MPI and MPI.COMM_WORLD.rank != rank:
        if rank >= MPI.COMM_WORLD.size:
            if MPI.COMM_WORLD.rank == 0:
                print("Specified rank (%d) is outside of the valid range (0-%d)." %
                      (rank, MPI.COMM_WORLD.size - 1))
            sys.exit()
        def color_print(s, **kwargs):
            pass
    else:
        def color_print(s, color='', end=''):
            """
            """
            print(color + s, file=stream, end='')
            print(Style.RESET_ALL, file=stream, end='')
            print(end=end)

    return color_print, Fore, Back, Style


def tree(top, show_solvers=True, show_jacs=True, show_colors=True, show_approx=True,
         filter=None, show_sizes=False, max_depth=0, rank=0, stream=sys.stdout):
    """
    Dump the model tree structure to the given stream.

    If you install colorama, the tree will be displayed in color if the stream is a terminal
    that supports color display.

    Parameters
    ----------
    top : System or Problem
        The top object in the tree.
    show_solvers : bool
        If True, include solver types in the tree.
    show_jacs : bool
        If True, include jacobian types in the tree.
    show_colors : bool
        If True and stream is a terminal that supports it, display in color.
    show_approx : bool
        If True, mark systems that are approximating their derivatives.
    filter : function(System)
        A function taking a System arg and returning None or an iter of (name, value) tuples.
        If None is returned, that system will not be displayed.  Otherwise, the system will
        be displayed along with any name, value pairs returned from the filter.
    show_sizes : bool
        If True, show input and output sizes for each System.
    max_depth : int
        Maximum depth for display.
    rank : int
        If MPI is active, the tree will only be displayed on this rank.  Only objects local
        to the given rank will be displayed.
    stream : File-like
        Where dump output will go.
    """
    from openmdao.core.problem import Problem
    from openmdao.core.group import Group
    from openmdao.core.implicitcomponent import ImplicitComponent

    cprint, Fore, Back, Style = _get_color_printer(stream, show_colors, rank=rank)

    tab = 0
    if isinstance(top, Problem):
        if filter is None:
            cprint('Driver: ', color=Fore.CYAN + Style.BRIGHT)
            cprint(type(top.driver).__name__, color=Fore.MAGENTA, end='\n')
            tab += 1
        top = top.model

    for s in top.system_iter(include_self=True, recurse=True):
        if filter is None:
            ret = ()
        else:
            ret = filter(s)
            if ret is None:
                continue

        depth = len(s.pathname.split('.')) if s.pathname else 0
        if max_depth != 0 and depth > max_depth:
            continue

        indent = '    ' * (depth + tab)
        cprint(indent, end='')

        if isinstance(s, Group):
            cprint("%s " % type(s).__name__, color=Fore.GREEN + Style.BRIGHT)
            cprint("%s" % s.name)
        else:
            if isinstance(s, ImplicitComponent):
                colr = Back.CYAN + Fore.BLACK + Style.BRIGHT
            else:
                colr = Fore.CYAN + Style.BRIGHT
            cprint("%s " % type(s).__name__, color=colr)
            cprint("%s" % s.name)
            if s._has_distrib_vars:
                cprint(" (distributed)", color=Fore.MAGENTA)

        # FIXME: these sizes could be wrong under MPI
        if show_sizes:
            cprint(" (%d / %d)" % (s._inputs._data.size, s._outputs._data.size),
                color=Fore.RED + Style.BRIGHT)

        if show_solvers:
            lnsolver = type(s.linear_solver).__name__
            nlsolver = type(s.nonlinear_solver).__name__

            if s.linear_solver is not None and lnsolver != "LinearRunOnce":
                cprint("  LN: ")
                cprint(lnsolver, color=Fore.MAGENTA + Style.BRIGHT)
            if s.nonlinear_solver is not None and nlsolver != "NonlinearRunOnce":
                cprint("  NL: ")
                cprint(nlsolver, color=Fore.MAGENTA + Style.BRIGHT)

        if show_jacs:
            jacs = []
            lnjac = nljac = None
            if s._assembled_jac is not None:
                lnjac = s._assembled_jac
                jacs.append(lnjac)
            if s.nonlinear_solver is not None:
                jacsolvers = list(s.nonlinear_solver._assembled_jac_solver_iter())
                if jacsolvers:
                    nljac = jacsolvers[0]._assembled_jac
                    if nljac is not lnjac:
                        jacs.append(nljac)

            if len(jacs) == 2:
                jnames = [' LN Jac: ', ' NL Jac: ']
            elif lnjac is not None:
                if lnjac is nljac:
                    jnames = [' Jac: ']
                else:
                    jnames = [' LN Jac: ']
            elif nljac is not None:
                jnames = [' NL Jac: ']
            else:
                jnames = []

            for jname, jac in zip(jnames, jacs):
                cprint(jname)
                cprint(type(jac).__name__, color=Fore.MAGENTA + Style.BRIGHT)

        if show_approx and s._approx_schemes:
            approx_keys = set()
            keys = set()
            for k, sjac in s._subjacs_info.items():
                if 'method' in sjac and sjac['method']:
                    approx_keys.add(k)
                else:
                    keys.add(k)
            diff = approx_keys - keys
            cprint("  APPROX: ", color=Fore.MAGENTA + Style.BRIGHT)
            cprint("%s (%d of %d)" % (list(s._approx_schemes), len(diff), len(s._subjacs_info)))

        cprint('', end='\n')

        vindent = indent + '  '
        for name, val in ret:
            cprint("%s%s: %s\n" % (vindent, name, val))


def _get_printer(comm, stream, rank=0):
    if comm.rank == rank:
        def p(*args, **kwargs):
            print(*args, file=stream, **kwargs)
    else:
        def p(*args, **kwargs):
            pass

    return p


def config_summary(problem, stream=sys.stdout):
    """
    Prints various high level statistics about the model structure.

    Parameters
    ----------
    problem : Problem
        The Problem to be summarized.
    stream : File-like
        Where the output will be written.
    """
    from openmdao.core.group import Group

    model = problem.model
    meta = model._var_allprocs_abs2meta
    locsystems = list(model.system_iter(recurse=True, include_self=True))
    locgroups = [s for s in locsystems if isinstance(s, Group)]

    grpnames = [s.pathname for s in locgroups]
    sysnames = [s.pathname for s in locsystems]
    ln_solvers = [(s.pathname, type(s.linear_solver).__name__) for s in locsystems
                              if s.linear_solver is not None]
    nl_solvers = [(s.pathname, type(s.nonlinear_solver).__name__) for s in locsystems
                         if s.nonlinear_solver is not None]

    max_depth = max([len(name.split('.')) for name in sysnames])
    setup_done = model._problem_meta['setup_status'] >= _SetupStatus.POST_FINAL_SETUP

    if problem.comm.size > 1:
        local_max = np.array([max_depth])
        global_max_depth = np.zeros(1, dtype=int)
        problem.comm.Allreduce(local_max, global_max_depth, op=MPI.MAX)

        proc_names = problem.comm.gather((sysnames, grpnames, ln_solvers, nl_solvers), root=0)
        grpnames = set()
        sysnames = set()
        ln_solvers = set()
        nl_solvers = set()
        if proc_names is not None:
            for systems, grps, lnsols, nlsols in proc_names:
                sysnames.update(systems)
                grpnames.update(grps)
                ln_solvers.update(lnsols)
                nl_solvers.update(nlsols)
    else:
        global_max_depth = max_depth
        ln_solvers = set(ln_solvers)
        nl_solvers = set(nl_solvers)

    ln_solvers = Counter([sname for _, sname in ln_solvers])
    nl_solvers = Counter([sname for _, sname in nl_solvers])

    # this gives us a printer that only prints on rank 0
    printer = _get_printer(problem.comm, stream)

    printer("============== Problem Summary ============")
    printer("Groups:           %5d" % len(grpnames))
    printer("Components:       %5d" % (len(sysnames) - len(grpnames)))
    printer("Max tree depth:   %5d" % global_max_depth)
    printer()

    if setup_done:
        desvars = model.get_design_vars()
        printer("Design variables:        %5d   Total size: %8d" %
                (len(desvars), sum(d['size'] for d in desvars.values())))

        con_nonlin_eq = {}
        con_nonlin_ineq = {}
        con_linear_eq = {}
        con_linear_ineq = {}
        for con, vals in model.get_constraints().items():
            if vals['linear']:
                if vals['equals'] is not None:
                    con_linear_eq[con] = vals
                else:
                    con_linear_ineq[con] = vals
            else:
                if vals['equals'] is not None:
                    con_nonlin_eq[con]= vals
                else:
                    con_nonlin_ineq[con]= vals

        con_nonlin = con_nonlin_eq.copy()
        con_nonlin.update(con_nonlin_ineq)
        con_linear = con_linear_eq.copy()
        con_linear.update(con_linear_ineq)

        printer("\nNonlinear Constraints:   %5d   Total size: %8d" %
                (len(con_nonlin), sum(d['size'] for d in con_nonlin.values())))
        printer("    equality:            %5d               %8d" %
                (len(con_nonlin_eq), sum(d['size'] for d in con_nonlin_eq.values())))
        printer("    inequality:          %5d               %8d" %
                (len(con_nonlin_ineq), sum(d['size'] for d in con_nonlin_ineq.values())))
        printer("\nLinear Constraints:      %5d   Total size: %8d" %
                (len(con_linear), sum(d['size'] for d in con_linear.values())))
        printer("    equality:            %5d               %8d" %
                (len(con_linear_eq), sum(d['size'] for d in con_linear_eq.values())))
        printer("    inequality:          %5d               %8d" %
                (len(con_linear_ineq), sum(d['size'] for d in con_linear_ineq.values())))

        objs = model.get_objectives()
        printer("\nObjectives:              %5d   Total size: %8d" %
                (len(objs), sum(d['size'] for d in objs.values())))

    printer()

    input_names = model._var_allprocs_abs2meta['input']
    ninputs = len(input_names)
    if setup_done:
        printer("Input variables:         %5d   Total size: %8d" %
                (ninputs, sum(meta['input'][n]['size'] for n in input_names)))
    else:
        printer("Input variables:         %5d" % ninputs)

    output_names = model._var_allprocs_abs2meta['output']
    noutputs = len(output_names)
    if setup_done:
        printer("Output variables:        %5d   Total size: %8d" %
                (noutputs, sum(meta['output'][n]['global_size'] for n in output_names)))
    else:
        printer("Output variables:        %5d" % noutputs)

    if setup_done and isinstance(model, Group):
        printer()
        conns = model._conn_global_abs_in2out
        printer("Total connections: %d   Total transfer data size: %d" %
                (len(conns), sum(meta['input'][n]['size'] for n in conns if n in meta['input'])))

    printer()
    printer("Driver type: %s" % problem.driver.__class__.__name__)
    linstr = []
    for slvname, num in ln_solvers.most_common():
        if num > 1:
            linstr.append('{} x {}'.format(slvname, num))
        else:
            linstr.append(slvname)
    printer("Linear Solvers: [{}]".format(', '.join(linstr)))


    nlstr = []
    for slvname, num in nl_solvers.most_common():
        if num > 1:
            nlstr.append('{} x {}'.format(slvname, num))
        else:
            nlstr.append(slvname)
    printer("Nonlinear Solvers: [{}]".format(', '.join(nlstr)))


def _summary_report(prob):
    path = prob.get_reports_dir() / 'summary.html'
    s = StringIO()
    config_summary(prob, s)
    with open(path, 'w') as f:
        f.write(text2html(s.getvalue()))


def _summary_report_register():
    register_report('summary', _summary_report, 'Model summary', 'Problem', 'final_setup', 'post')


@contextmanager
def profiling(outname='prof.out'):
    """
    Context manager that runs cProfile on the wrapped code and dumps stats to the given filename.

    Parameters
    ----------
    outname : str
        Name of the output file containing profiling stats.
    """
    import cProfile
    prof = cProfile.Profile()
    prof.enable()

    try:
        yield prof
    finally:
        prof.disable()
        prof.dump_stats(outname)


def compare_jacs(Jref, J, rel_trigger=1.0):
    results = []

    for key in set(J).union(Jref):
        if key in J:
            subJ = J[key]
        else:
            subJ = np.zeros(Jref[key].shape)

        if key in Jref:
            subJref = Jref[key]
        else:
            subJref = np.zeros(J[key].shape)

        diff = np.abs(subJ - subJref)
        absref = np.abs(subJref)
        rel_idxs = np.nonzero(absref > rel_trigger)
        diff[rel_idxs] /= absref[rel_idxs]

        max_diff_idx = np.argmax(diff)
        max_diff = diff.flatten()[max_diff_idx]

        # now determine if max diff is abs or rel
        diff[:] = 0.0
        diff[rel_idxs] = 1.0
        if diff.flatten()[max_diff_idx] > 0.0:
            results.append((key, max_diff, 'rel'))
        else:
            results.append((key, max_diff, 'abs'))

    return results


def trace_dump(fname='trace_dump', skip=(), flush=True):
    """
    Dump traces to the specified filename<.rank> showing openmdao and c calls.

    Under MPI it will write a separate file for each rank.

    Parameters
    ----------
    fname : str
        Name of the trace file(s).  <.rank> will be appended to the name on each rank.
    skip : set-like
        Collection of function names to skip.
    flush : bool
        If True, flush print buffer after every print call.
    """
    if sys.getprofile() is not None:
        raise RuntimeError("another profile function is already active.")

    suffix = '.0' if MPI is None else '.' + str(MPI.COMM_WORLD.rank)
    my_fname = fname + suffix

    outfile = open(my_fname, 'w')

    stack = []

    _c_map = {
        'c_call': '(c) -->',
        'c_return': '(c) <--',
        'c_exception': '(c_exception)',
    }


    def _print_c_func(frame, arg, typestr):
        s = str(arg)
        if 'mpi4py' in s or 'petsc4py' in s:
            c = arg.__self__.__class__
            if stack:
                pname = f"(scope: {stack[-1][0]})"
            else:
                pname = ''
            print('   ' * len(stack), typestr, f"{c.__module__}.{c.__name__}.{arg.__name__}",
                  f"{frame.f_code.co_filename}:{frame.f_code.co_firstlineno} {pname}",
                  file=outfile, flush=True)


    def _mpi_trace_callback(frame, event, arg):
        pname = None
        commsize = ''
        if event == 'call':
            if 'openmdao' in frame.f_code.co_filename:
                if frame.f_code.co_name in skip:
                    return
                if 'self' in frame.f_locals:
                    try:
                        pname = frame.f_locals['self'].msginfo
                    except Exception:
                        pass
                    try:
                        commsize = frame.f_locals['self'].comm.size
                    except Exception:
                        pass
                if pname is not None:
                    if not stack or pname != stack[-1][0]:
                        stack.append([pname, 1])
                        print('   ' * len(stack), commsize, f"(scope: {pname})", file=outfile, flush=flush)
                    else:
                        stack[-1][1] += 1
                print('   ' * len(stack), '-->', frame.f_code.co_name, "%s:%d" %
                      (frame.f_code.co_filename, frame.f_code.co_firstlineno),
                      file=outfile, flush=flush)
        elif event == 'return':
            if 'openmdao' in frame.f_code.co_filename:
                if frame.f_code.co_name in skip:
                    return
                if 'self' in frame.f_locals:
                    try:
                        pname = frame.f_locals['self'].msginfo
                    except Exception:
                        pass
                    try:
                        commsize = frame.f_locals['self'].comm.size
                    except Exception:
                        pass
                print('   ' * len(stack), '<--', frame.f_code.co_name, "%s:%d" %
                      (frame.f_code.co_filename, frame.f_code.co_firstlineno),
                      file=outfile, flush=flush)
                if pname is not None and stack and pname == stack[-1][0]:
                    stack[-1][1] -= 1
                    if stack[-1][1] < 1:
                        stack.pop()
                        if stack:
                            print('   ' * len(stack), commsize, stack[-1][0], file=outfile,
                                  flush=flush)
        else:
            _print_c_func(frame, arg, _c_map[event])

    sys.setprofile(_mpi_trace_callback)


def prom_info_dump(system, tgt):
    """
    Dump the promotion src_indices/src_shape data for the given absolute target name.

    The data actually lives in the Problem metadata, but is more convenient to access during
    debugging by using a System instance to access that metadata.

    Promotion src_indices/src_shape data is displayed for all inputs, including tgt, that
    are connected to the same source.

    Parameters
    ----------
    system : System
        Any System instance.
    tgt : str
        Absolute name of an input variable.
    """
    probmeta = system._problem_meta
    model = probmeta['model_ref']()
    src = model._conn_global_abs_in2out[tgt]
    abs_in2prom_info = probmeta['abs_in2prom_info']
    print('For tgt', tgt, 'and src', src, 'connected tgts and prom info are:')
    for t, s in model._conn_global_abs_in2out.items():
        if s == src:
            print('    ', t)
            if t in abs_in2prom_info:
                for p in abs_in2prom_info[t]:
                    print('        ', p)
    print(flush=True)


def comm_info(system, outfile=None, verbose=False, table_format='box_grid'):
    """
    Display MPI communicator information for Systems in the model.

    Parameters
    ----------
    system : System
        A System in the model.
    outfile : str or None
        Name of file where output will be written. If None, output is written to stdout.
    verbose : bool
        If True, display comm size and rank range for all Systems. Otherwise, display only Systems
        having a comm size different from their parent system.
    table_format : str
        Table format.  All formats compatible with the generate_table function are available.
    """
    if MPI and MPI.COMM_WORLD.size > 1:
        dct = {}
        for path, csize, rank, wrank in system.comm_info_iter():
            if path not in dct:
                dct[path] = [csize, wrank, wrank]
            else:
                csize, minwrnk, maxwrnk = dct[path]
                # do min/max here *and* after the gather so we don't have to
                # gather all the data to get the min/max
                if wrank < minwrnk:
                    minwrnk = wrank
                if wrank > maxwrnk:
                    maxwrnk = wrank
                dct[path] = [csize, minwrnk, maxwrnk]

        # collect dct from all procs so we can get the full min/max rank range
        alldcts = system.comm.gather(dct, root=0)

        if MPI.COMM_WORLD.rank == 0:
            alldata = {}
            sizes = {}
            for dct in alldcts:
                for path, v in dct.items():
                    csize, newmin, newmax = v
                    sizes[path] = csize
                    if path in alldata:
                        _, minwrnk, maxwrnk = alldata[path]
                        if newmin < minwrnk:
                            minwrnk = newmin
                        if newmax > maxwrnk:
                            maxwrnk = newmax
                        alldata[path] = [csize, minwrnk, maxwrnk]
                    else:
                        alldata[path] = [csize, newmin, newmax]

            table_data = []
            headers = ['Comm Size', 'COMM_WORLD Range(s)', 'System Pathname']
            for path, lst in sorted(alldata.items(), key=lambda x: (-x[1][0], x[0])):
                csize, minwrnk, maxwrnk = lst
                if verbose or path == '' or sizes[path.rpartition('.')[0]] != csize:
                    if minwrnk == maxwrnk:
                        rng = str(minwrnk)
                    else:
                        rng = f"{minwrnk} - {maxwrnk}"
                    table_data.append([csize, rng, path])

            col_meta = [
                {'align': 'center', 'header_align': 'center'},
                {'align': 'center', 'header_align': 'center'},
                {}
            ]
            if table_format != 'tabulator':
                col_meta[0]['max_width'] = 5
                col_meta[1]['max_width'] = 15

            probpath = system._problem_meta['pathname']
            if outfile is None:
                print(f"Printing comm info table for Problem '{probpath}'")

            outf = generate_table(table_data, headers=headers, column_meta=col_meta,
                                  tablefmt=table_format).display(outfile=outfile)

            if outf is not None and MPI.COMM_WORLD.rank == 0:
                print(f"comm info table for Problem '{probpath}' written to {outf}")
    else:
        if outfile is None:
            print("No MPI process info available.")
        else:
            with open(outfile, 'w') as f:
                print("No MPI process info available.", file=f)


def is_full_slice(range, inds):
    size = range[1] - range[0]
    inds = np.asarray(inds)
    if len(inds) > 1 and inds[0] == 0 and inds[-1] == size - 1:
        step = inds[1] - inds[0]
        diffs = inds[1:] - inds[:-1]
        return np.all(diffs == step)

    return len(inds) == 1 and inds[0] == 0


def show_dist_var_conns(group, rev=False, out_stream=_DEFAULT_OUT_STREAM):
    """
    Show all distributed variable connections in the given group and below.

    The ranks displayed will be relative to the communicator of the given top group.

    Parameters
    ----------
    group : Group
        The top level group to be searched.  Connections in all subgroups will also be displayed.
    rev : bool
        If True show reverse transfers instead of forward transfers.
    out_stream : file-like
        Where the output will go.

    Returns
    -------
    dict or None
        Dictionary containing the data for the connections. None is returned on all ranks except
        rank 0.
    """
    from openmdao.core.group import Group

    if out_stream is _DEFAULT_OUT_STREAM:
        out_stream = sys.stdout

    if out_stream is None:
        def printer(*args, **kwargs):
            return None
    else:
        printer = print

    direction = 'rev' if rev else 'fwd'
    arrowdir = {'fwd': '->', 'rev': '<-'}
    arrow = arrowdir[direction]

    gdict = {}

    for g in group.system_iter(typ=Group, include_self=True):
        if g._transfers[direction]:
            in_ranges = list(g.dist_size_iter('input', group.comm))
            out_ranges = list(g.dist_size_iter('output', group.comm))

            inmapper = RangeMapper.create(in_ranges)
            outmapper = RangeMapper.create(out_ranges)

            gprint = False

            gprefix = g.pathname + '.' if g.pathname else ''
            skip = len(gprefix)

            for sub, transfer in g._transfers[direction].items():
                if sub is not None and (not isinstance(sub, tuple) or sub[0] is not None):
                    if not gprint:
                        gdict[g.pathname] = {}
                        gprint = True

                    conns = {}
                    for iidx, oidx in zip(transfer._in_inds, transfer._out_inds):
                        idata, irind = inmapper.index2key_rel(iidx)
                        ivar, irank = idata
                        odata, orind = outmapper.index2key_rel(oidx)
                        ovar, orank = odata

                        if odata not in conns:
                            conns[odata] = {}
                        if idata not in conns[odata]:
                            conns[odata][idata] = []
                        conns[odata][idata].append((orind, irind))

                    strs = {}  # use to avoid duplicate printouts for different ranks

                    for odata, odict in conns.items():
                        ovar, orank = odata
                        ovar = ovar[skip:]

                        for idata, dlist in odict.items():
                            ivar, irank = idata
                            ivar = ivar[skip:]
                            ranktup = (orank, irank)

                            oinds = [d[0] for d in dlist]
                            iinds = [d[1] for d in dlist]

                            orange = outmapper.key2range(odata)
                            irange = inmapper.key2range(idata)

                            if is_full_slice(orange, oinds) and is_full_slice(irange, iinds):
                                s = f"{ovar} {arrow} {ivar}"
                                if s not in strs:
                                    strs[s] = set()
                                strs[s].add(ranktup)
                                continue

                            for oidx, iidx in zip(oinds, iinds):
                                s = f"{ovar}[{oidx}] {arrow} {ivar}[{iidx}]"
                                if s not in strs:
                                    strs[s] = set()
                                strs[s].add(ranktup)

                    gdict[g.pathname][str(sub)] = strs

    do_ranks = False

    if group.comm.size > 1:
        do_ranks = True
        final = {}
        gatherlist = group.comm.gather(gdict, root=0)
        if group.comm.rank == 0:
            for dct in gatherlist:
                for gpath, subdct in dct.items():
                    if gpath not in final:
                        final[gpath] = subdct
                    else:
                        fgpath = final[gpath]
                        for sub, strs in subdct.items():
                            if sub not in fgpath:
                                fgpath[sub] = strs
                            else:
                                fgpathsub = fgpath[sub]
                                for s, ranks in strs.items():
                                    if s not in fgpathsub:
                                        fgpathsub[s] = ranks
                                    else:
                                        fgpathsub[s] |= ranks

        gdict = final

    if group.comm.rank == 0:
        fwd = direction == 'fwd'
        for gpath, subdct in sorted(gdict.items(), key=lambda x: x[0]):
            indent = 0 if gpath == '' else gpath.count('.') + 1
            pad = '   ' * indent
            printer(f"{pad}In Group '{gpath}'", file=out_stream)
            for sub, strs in sorted(subdct.items(), key=lambda x: x[0]):
                if fwd:
                    printer(f"{pad}   {arrow} {sub}", file=out_stream)
                else:
                    printer(f"{pad}   {sub} {arrow}", file=out_stream)
                for s, ranks in strs.items():
                    if do_ranks:
                        oranks = np.empty(len(ranks), dtype=int)
                        iranks = np.empty(len(ranks), dtype=int)
                        for i, (ornk, irnk) in enumerate(sorted(ranks)):
                            oranks[i] = ornk
                            iranks[i] = irnk

                        if np.all(oranks == oranks[0]):
                            orstr = str(oranks[0])
                        else:
                            sorted_ranks = sorted(oranks)
                            orstr = str(sorted_ranks)
                            if len(sorted_ranks) > 3:
                                for j, r in enumerate(sorted_ranks):
                                    if j == 0 or r - val == 1:  # noqa: F821, val initialized below
                                        val = r
                                    else:
                                        break
                                else:
                                    orstr = f"[{sorted_ranks[0]} to {sorted_ranks[-1]}]"

                        if np.all(iranks == iranks[0]):
                            irstr = str(iranks[0])
                        else:
                            sorted_ranks = sorted(iranks)
                            irstr = str(sorted(iranks))
                            if len(sorted_ranks) > 3:
                                for j, r in enumerate(sorted_ranks):
                                    if j == 0 or r - val == 1:  # noqa: F821, val initialized below
                                        val = r
                                    else:
                                        break
                                else:
                                    irstr = f"[{sorted_ranks[0]} to {sorted_ranks[-1]}]"

                        if orstr == irstr and '[' not in orstr:
                            printer(f"{pad}      {s}    rank {orstr}", file=out_stream)
                        else:
                            printer(f"{pad}      {s}    ranks {orstr} {arrow} {irstr}", file=out_stream)
                    else:
                        printer(f"{pad}      {s}", file=out_stream)

        return gdict


def _dist_conns_setup_parser(parser):
    """
    Set up the openmdao subparser for the 'openmdao dist_conns' command.

    Parameters
    ----------
    parser : argparse subparser
        The parser we're adding options to.
    """
    parser.add_argument('file', nargs=1, help='Python file containing the model.')
    parser.add_argument('-o', default=None, action='store', dest='outfile',
                        help='Name of output file.  By default, output goes to stdout.')
    parser.add_argument('-r', '--rev', action='store_true', dest='rev',
                        help='If set, use "rev" transfer direction instead of "fwd".')
    parser.add_argument('-p', '--problem', action='store', dest='problem', help='Problem name')


def _dist_conns_cmd(options, user_args):
    """
    Run the `openmdao dist_conns` command.

    The command shows connections, with relative indexing information, between all
    variables in the model across all MPI processes.

    Parameters
    ----------
    options : argparse Namespace
        Command line options.
    user_args : list of str
        Args to be passed to the user script.
    """
    import openmdao.utils.hooks as hooks

    def _dist_conns(prob):
        model = prob.model
        if options.problem:
            if model._problem_meta['name'] != options.problem and \
                    model._problem_meta['pathname'] != options.problem:
                return
        elif '/' in model._problem_meta['pathname']:
            # by default, only display comm info for a top level problem
            return

        if options.outfile is None:
            out_stream = sys.stdout
        else:
            out_stream = open(options.outfile, 'w')

        try:
            show_dist_var_conns(model, rev=options.rev, out_stream=out_stream)
        finally:
            if out_stream is not sys.stdout:
                out_stream.close()

        exit()

    # register the hook to be called right after final_setup on the problem
    hooks._register_hook('final_setup', 'Problem', post=_dist_conns)

    _load_and_exec(options.file[0], user_args)



import unittest
import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_totals
from openmdao.utils.testing_utils import use_tempdirs

class Inner(om.Group):
    def setup(self):
        comp = om.ExecComp('y=2*x', x=np.zeros((3, 2)), y=np.zeros((3, 2)))
        self.add_subsystem('comp', comp)


class Outer(om.Group):
    def setup(self):
        self.add_subsystem('inner', Inner())
    def configure(self):
        self.promotes('inner', inputs=[('comp.x', 'desvar_x')],
                      src_indices=np.array([0, 1, 0, 1, 0, 1]), flat_src_indices=True, src_shape=2)


class SrcIndicesTestCase(unittest.TestCase):
    def test_one_nesting(self):
        prob = om.Problem()
        model = prob.model
        model.add_subsystem('src', om.ExecComp('y=3*x', x=np.zeros((7)), y=np.zeros((7))))
        model.add_subsystem('outer', Outer())
        model.connect('src.y', 'outer.desvar_x', src_indices=[2, 4], flat_src_indices=True)
        prob.setup()
        srcval = np.array([1.0, 3.0, 5.0, 7.0, 9.0, 11.0, 13.0])
        prob.set_val('src.x', srcval)
        prob.run_model()
        assert_near_equal(prob.get_val('outer.desvar_x'), [15., 27.], 1e-6)
        expected = np.array([[15., 27.],
                             [15., 27.],
                             [15., 27.]])
        assert_near_equal(prob.get_val('outer.inner.comp.x'), expected, 1e-6)

    def test_broadcast_scalar_connection(self):
        """
        OpenMDAO should allow promotion to an input that appears to have a certain shape.
        Users should be allowed to connect an arbitrary variable with compatible src_indices to that input.
        """
        p = om.Problem()

        # c1 contains scalar calculations
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp())
        ivc.add_output('y', val=5.0*np.ones(4,))
        ivc.add_output('x', val=np.arange(4, dtype=float) + 1)

        g1 = p.model.add_subsystem('g1', om.Group())
        # c2 is vectorized calculations
        g1.add_subsystem('c2', om.ExecComp('z = a * y', shape=(4,)))

        # The ultimate source of a and y may be scalar, or have some other arbitrary shape
        g1.promotes('c2', inputs=['a'], src_indices=[0, 0, 0, 0], src_shape=(1,))
        g1.promotes('c2', inputs=['y'], src_indices=[0, 0, 0, 0], src_shape=(4,))

        p.model.connect('ivc.y', 'g1.y')

        # Now connect only a portion of some other output to a, which appears as a scalar input
        p.model.connect('ivc.x', 'g1.a', src_indices=[-1])

        p.setup()

        p.run_model()

        assert_near_equal(p['g1.a'], 4.)
        assert_near_equal(p.model.g1.get_val('a'), 4.)
        assert_near_equal(p['g1.y'], [5., 5., 5., 5.])
        assert_near_equal(p.model.g1.get_val('y'), [5., 5., 5., 5.])
        assert_near_equal(p['g1.c2.y'], [5., 5., 5., 5.])
        assert_near_equal(p['g1.c2.a'], [4., 4., 4., 4.])
        assert_near_equal(p['g1.c2.z'], [20., 20., 20., 20.])

    def test_multiple_inputs_different_src_indices(self):
        """
        When different variables get promoted to the same name, but have different src_indices, this
        should be supported as long as the src_shape of the different promotions is compatible or
        made so by the set_input_defaults call.
        """
        p = om.Problem()

        g1 = p.model.add_subsystem('g1', om.Group(), promotes_inputs=['b'])
        # c1 contains scalar calculations
        g1.add_subsystem('c1', om.ExecComp('y = a0 + b', shape=(1,)),
                         promotes_inputs=[('a0', 'a'), 'b'], promotes_outputs=['y'])
        g2 = g1.add_subsystem('g2', om.Group())
        # c2 is vectorized calculations
        g2.add_subsystem('c2', om.ExecComp('z = a * y', shape=(4,)), promotes_inputs=['a', 'y'])

        g1.promotes('g2', inputs=['y'], src_indices=[0, 0, 0, 0], src_shape=(1,))
        g1.promotes('g2', inputs=['a'], src_indices=[0, 0, 0, 0], src_shape=(1,))

        p.model.promotes('g1', inputs=['a'], src_indices=[0], src_shape=(1,))

        p.setup()

        p['a'] = 99
        p['b'] = 2

        p.run_model()

        assert_near_equal(p['a'], 99.)
        assert_near_equal(p['b'], 2.)
        assert_near_equal(p.model.get_val('a'), 99.)
        assert_near_equal(p.model.get_val('b'), 2.)
        assert_near_equal(p['g1.y'], 101.)
        assert_near_equal(p['g1.a'], 99.)
        assert_near_equal(p.model.g1.get_val('y'), 101.)
        assert_near_equal(p.model.g1.get_val('a'), 99.)
        assert_near_equal(p['g1.c1.a0'], 99.)
        assert_near_equal(p['g1.c1.b'], 2.)
        assert_near_equal(p['g1.g2.y'], [101.] * 4)
        assert_near_equal(p.model.g1.g2.get_val('y'), [101.] * 4)
        assert_near_equal(p['g1.g2.a'], [99.] * 4)
        assert_near_equal(p.model.g1.g2.get_val('a'), [99.] * 4)
        assert_near_equal(p['g1.g2.c2.a'], [99.] * 4)
        assert_near_equal(p['g1.g2.c2.y'], [101.] * 4)
        assert_near_equal(p['g1.g2.c2.z'], [9999.] * 4)

    def test_src_indices_nested_promotes(self):
        """
        Promoting a variable up multiple levels with different src_indices at each level.
        """
        p = om.Problem()

        g1 = p.model.add_subsystem('g1', om.Group(), promotes_inputs=['b'])
        # c1 contains scalar calculations
        g1.add_subsystem('c1', om.ExecComp('y = a0 + b', shape=(1,)),
                         promotes_inputs=[('a0', 'a'), 'b'], promotes_outputs=['y'])

        g2 = g1.add_subsystem('g2', om.Group())
        # c2 is vectorized calculations
        g2.add_subsystem('c2',  om.ExecComp('z = a * y', shape=(4,)), promotes_inputs=['a', 'y'])

        g1.promotes('g2', inputs=['a'], src_indices=[0, 0, 0, 0], src_shape=(1,))
        g1.promotes('g2', inputs=['y'], src_indices=[0, 0, 0, 0], src_shape=(1,))

        p.model.promotes('g1', inputs=['a'], src_indices=[0], src_shape=(1,))

        p.setup()

        p['a'] = 99
        p['b'] = 2

        p.run_model()

        assert_near_equal(p['g1.y'], 101.)
        assert_near_equal(p['g1.g2.c2.z'], [9999.] * 4)

    def test_mixed_src_indices_no_src_indices(self):
        class C1(om.ExplicitComponent):
            def setup(self):
                self.add_input('diameter', 0.0, units='m')
                self.add_output('z_start', 0.0, units='m')

            def compute(self, inputs, outputs):
                outputs['z_start'] = inputs['diameter'] * 2.


        class C2(om.ExplicitComponent):

            def setup(self):
                self.add_input('diameter', np.zeros(3), units='m')

            def compute(self, inputs, outputs):
                pass

        # C1 has src_indices and C2 doesn't.
        prob = om.Problem()
        prob.model.add_subsystem('C1', C1())
        prob.model.promotes('C1', inputs=['diameter'], src_indices=[0])
        prob.model.add_subsystem('C2', C2(), promotes=['diameter'])  # size 3

        prob.setup()

        prob['diameter'] = np.ones(3) * 1.5

        prob.run_model()

        assert_near_equal(prob['C1.z_start'], 3.)

        assert_near_equal(prob['diameter'], np.ones(3) * 1.5)
        assert_near_equal(prob['C1.diameter'], 1.5)
        assert_near_equal(prob['C2.diameter'], np.ones(3) * 1.5)

    def test_flat_src_inds_2_levels(self):
        class Burn1(om.Group):
            def setup(self):
                self.add_subsystem('comp1', om.ExecComp(['y1=x*2'], y1=np.ones(4), x=np.ones(4)),
                                   promotes_outputs=['*'])

                self.add_subsystem('comp2', om.ExecComp(['y2=x*3'], y2=np.ones(4), x=np.ones(4)),
                                   promotes_outputs=['*'])

            def configure(self):
                self.promotes('comp1', inputs=[('x', 'design:x')],
                              src_indices=[0, 0, 0, 0], flat_src_indices=True)

                self.set_input_defaults('design:x', 75.3)


        class Traj(om.Group):
            def setup(self):
                self.add_subsystem('burn1', Burn1(), promotes_outputs=['*'])

            def configure(self):
                self.promotes('burn1', inputs=['design:x'],
                              src_indices=[0, 0, 0, 0], flat_src_indices=True)

        prob = om.Problem(model=Traj())

        prob.setup()
        prob.run_model()

        assert_near_equal(prob['design:x'], [75.3])
        assert_near_equal(prob['y1'], [75.3*2]*4)
        assert_near_equal(prob['y2'], [1*3]*4)

    def test_src_inds_2_subs(self):

        class RHS(om.Group):

            def initialize(self):
                self.options.declare('size', 1)

            def setup(self):
                size = self.options['size']
                self.add_subsystem('comp1', om.ExecComp(['y1=x*2'], y1=np.ones(size), x=np.ones(size)),
                                promotes_inputs=['*'], promotes_outputs=['*'])

                # test with second absolute path for 'x'
                self.add_subsystem('comp2', om.ExecComp(['y2=x*3'], y2=np.ones(size), x=np.ones(size)),
                                promotes_inputs=['*'], promotes_outputs=['*'])


        class Phase(om.Group):
            def setup(self):
                self.add_subsystem('rhs', RHS(size=4))

            def configure(self):
                self.promotes('rhs', inputs=[('x', 'design:x')],
                            src_indices=[0, 0, 0, 0], flat_src_indices=True)

                # this doesn't set the value because it's connected to a non-auto_ivc that
                # already has a value of 1.0 that overrides it.
                self.set_input_defaults('design:x', 75.3)


        class Traj(om.Group):
            def setup(self):
                self.add_subsystem('src', om.ExecComp(['q = b*3'], b=1.5))

                self.add_subsystem('phase', Phase())

                self.connect('src.q', 'phase.design:x')

        prob = om.Problem(model=Traj())

        prob.setup()
        prob.run_model()

        assert_near_equal(prob['phase.design:x'], 4.5)
        assert_near_equal(prob['phase.rhs.y1'], [4.5*2]*4)
        assert_near_equal(prob['phase.rhs.y2'], [4.5*3]*4)

    def test_sub_sub_promotes(self):
        class Burn1(om.Group):
            def setup(self):
                self.add_subsystem('comp1', om.ExecComp(['y1=x*2'], y1=np.ones(4), x=np.ones(4)),
                                promotes_outputs=['*'])
                self.add_subsystem('comp2', om.ExecComp(['y2=x*2'], y2=np.ones(4), x=np.ones(4)),
                                promotes_outputs=['*'])

            def configure(self):
                self.promotes('comp1', inputs=[('x', 'design:x')], src_indices=[0, 0, 0, 0], flat_src_indices=True)
                self.set_input_defaults('design:x', 75.3)

        class Phases(om.ParallelGroup):
            def setup(self):
                self.add_subsystem('burn1', Burn1(), promotes_outputs=['*'])

        class Traj(om.Group):
            def setup(self):
                self.add_subsystem('phases', Phases(), promotes_outputs=['*'])

            def configure(self):
                # this promotes was leaving a leftover entry in _group_inputs that resulted
                # from an earlier _setup_var_data call where the input in question had a different
                # promoted name.
                self.phases.promotes('burn1', inputs=['design:x'])

        prob = om.Problem(model=Traj())
        prob.setup()
        prob.run_model()

        assert_near_equal(prob['phases.design:x'], 75.3)
        assert_near_equal(prob['y1'], [75.3*2]*4)

    def test_src_shape_mismatch(self):
        p = om.Problem(name='src_shape_mismatch')
        G = p.model.add_subsystem('G', om.Group(), promotes_inputs=['x'])

        G.set_input_defaults('x', src_shape=(3,2))

        g1 = G.add_subsystem('g1', om.Group(), promotes_inputs=['x'])
        g1.add_subsystem('C1', om.ExecComp('y = 3*x', shape=3))

        g1.promotes('C1', inputs=['x'], src_indices=om.slicer[:, 1], src_shape=(3,3))

        g2 = G.add_subsystem('g2', om.Group(), promotes_inputs=['x'])
        g2.add_subsystem('C2', om.ExecComp('y = 2*x', shape=2))

        g2.promotes('C2', inputs=['x'], src_indices=[1,5], src_shape=(3,2), flat_src_indices=True)

        with self.assertRaises(Exception) as cm:
            p.setup()

        self.assertEqual(cm.exception.args[0],
           "\nCollected errors for problem 'src_shape_mismatch':"
           "\n   <model> <class Group>: When connecting '_auto_ivc.v0' to 'G.g1.C1.x': Promoted "
           "src_shape of (3, 3) for 'G.g1.C1.x' differs from src_shape (3, 2) for 'x'.")

    def test_src_indices_on_promotes(self):
        src_shape = (3, 3)
        tgt_shape = (2, 2)
        src_indices = [4, 5, 7, 9]
        flat_src_indices = True

        class MyComp(om.ExplicitComponent):
            def __init__(self, input_shape):
                super().__init__()
                self._input_shape = input_shape
            def setup(self):
                self.add_input('x', val=np.zeros(self._input_shape))
                self.add_output('y', val=np.zeros(self._input_shape))
            def compute(self, inputs, outputs):
                outputs['y'] = 2.0 * inputs['x']

        p = om.Problem(name='src_indices_on_promotes')
        p.model.add_subsystem('indeps', om.IndepVarComp('x', shape=src_shape))
        p.model.add_subsystem('C1', MyComp(tgt_shape))
        p.model.promotes('C1', any=['x'],
                            src_indices=src_indices,
                            flat_src_indices=flat_src_indices)
        p.model.set_input_defaults('x', src_shape=src_shape)
        with self.assertRaises(Exception) as cm:
            p.setup()

        self.assertEqual(cm.exception.args[0],
            "\nCollected errors for problem 'src_indices_on_promotes':"
            "\n   <model> <class Group>: When promoting 'x' from system 'C1' with src_indices "
            "[4 5 7 9] and src_shape (3, 3): index 9 is out of bounds for source dimension of size 9."
            "\n   <model> <class Group>: The source indices [4 5 7 9] do not specify a valid shape "
            "for the connection '_auto_ivc.v0' to 'C1.x'. (target shape=(2, 2), indices_shape=(4,)):"
            " index 9 is out of bounds for axis 0 with size 9")

    def test_connect_slice_src_indices_not_full_size(self):
        p = om.Problem()

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp())
        M_np = np.arange(9)[::-1].reshape((3, 3))
        ivc.add_output('M', val=M_np)

        exec = p.model.add_subsystem('exec', om.ExecComp())

        exec.add_expr('A = B', A={'shape': (3,)}, B={'shape': (3,)})

        p.model.connect('ivc.M', 'exec.B', src_indices=om.slicer[:1])

        p.setup()

        p.run_model()

        assert_near_equal(p.get_val('exec.B').ravel(), p.get_val('ivc.M')[:1].ravel())
        assert_near_equal(p.get_val('exec.A').ravel(), p.get_val('ivc.M')[:1].ravel())

    def test_promote_slice_src_indices_not_full_size(self):
        p = om.Problem()

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp())
        M_np = np.arange(9)[::-1].reshape((3, 3))
        ivc.add_output('M', val=M_np)

        exec = p.model.add_subsystem('exec', om.ExecComp())

        exec.add_expr('A = B', A={'shape': (3,)}, B={'shape': (3,)})

        p.model.promotes('ivc', outputs=['M'])
        p.model.promotes('exec', inputs=[('B', 'M')], src_indices=om.slicer[:1], src_shape=(3, 3))

        p.setup()

        p.run_model()

        assert_near_equal(p.get_val('exec.B').ravel(), p.get_val('ivc.M')[:1].ravel())
        assert_near_equal(p.get_val('exec.A').ravel(), p.get_val('ivc.M')[:1].ravel())

class SrcIndicesFeatureTestCase(unittest.TestCase):
    def test_multi_promotes(self):

        p = om.Problem()
        G = p.model.add_subsystem('G', om.Group())

        # At the top level, we assume that the source, and our input 'x', has a shape of (3,3),
        # and after we slice it with [:,:-1], lower levels will see their source having a shape of (3,2)
        p.model.promotes('G', inputs=['x'], src_indices=om.slicer[:,:-1], src_shape=(3,3))

        # This specifies that G.x assumes a source shape of (3,2)
        G.set_input_defaults('x', src_shape=(3,2))

        g1 = G.add_subsystem('g1', om.Group(), promotes_inputs=['x'])
        g1.add_subsystem('C1', om.ExecComp('y = 3*x', x=np.random.random(3), y=np.random.random(3)))

        # C1.x has a shape of 3, so we apply a slice of [:, 1] to our source which has a shape
        # of (3,2) to give us our final shape of 3.
        g1.promotes('C1', inputs=['x'], src_indices=om.slicer[:, 1], src_shape=(3,2))

        g2 = G.add_subsystem('g2', om.Group(), promotes_inputs=['x'])
        g2.add_subsystem('C2', om.ExecComp('y = 2*x', shape=2))

        # C2.x has a shape of 2, so we apply flat source indices of [1,5] to our source which has
        # a shape of (3,2) to give us our final shape of 2.
        g2.promotes('C2', inputs=['x'], src_indices=[1,5], flat_src_indices=True, src_shape=(3,2))

        p.setup()

        inp = np.arange(9).reshape((3,3)) + 1.

        # 'x' at the top level has a shape of (3,3) because no src_indices have been applied yet
        p.set_val('x', inp)
        p.run_model()

        assert_near_equal(p['x'], inp)
        assert_near_equal(p['G.x'], inp[:, :-1])
        assert_near_equal(p['G.g1.C1.y'], inp[:, :-1][:, 1]*3.)
        assert_near_equal(p['G.g2.C2.y'], inp[:, :-1].flatten()[[1,5]]*2.)


@use_tempdirs
class SrcIndicesMPITestCase(unittest.TestCase):
    N_PROCS = 2

    def test_multi_promotes_mpi(self):
        p = om.Problem()
        par = p.model.add_subsystem('par', om.ParallelGroup())
        g1 = par.add_subsystem('g1', om.Group(), promotes_inputs=['x'])
        g2 = par.add_subsystem('g2', om.Group(), promotes_inputs=['x'])
        g1.add_subsystem('C1', om.ExecComp('y = 3*x', shape=3))
        g2.add_subsystem('C2', om.ExecComp('y = 2*x', shape=2))
        g1.promotes('C1', inputs=['x'], src_indices=om.slicer[:, 1], src_shape=(3,2))
        g2.promotes('C2', inputs=['x'], src_indices=[1,5], src_shape=(3,2), flat_src_indices=True)

        # we want the connection to x to have a shape of (3,2), which differs from the
        # shapes of either of the connected absolute inputs.
        par.set_input_defaults('x', src_shape=(3,2))

        # we want the auto_ivc output to have a shape of (3,3)
        p.model.promotes('par', inputs=['x'], src_indices=om.slicer[:,:-1], src_shape=(3,3))

        p.setup()

        commsize = p.comm.size
        inp = np.random.random((3,3))
        if commsize > 1:
            if p.comm.rank == 0:
                p.comm.bcast(inp, root=0)
            else:
                inp = p.comm.bcast(None, root=0)

        reduced_inp = inp[:, :-1]

        p.set_val('x', inp)
        p.run_model()

        if commsize == 1 or p.comm.rank == 0:
            assert_near_equal(p['par.g1.C1.y'], reduced_inp[:, 1]*3.)
        elif commsize == 1 or p.comm.rank == 1:
            assert_near_equal(p['par.g2.C2.y'], reduced_inp.flatten()[[1,5]]*2.)


class SrcIndicesSerialMultipoint2(unittest.TestCase):

    def test_multipoint2(self):
        p = om.Problem()
        par = p.model.add_subsystem('par', om.ParallelGroup())

        g1 = par.add_subsystem('g1', om.Group(), promotes_inputs=['x'])
        g1.add_subsystem('C1', om.ExecComp('y = 3*x', shape=1))
        g1.promotes('C1', inputs=['x'], src_indices=[0], src_shape=(2,))

        g2 = par.add_subsystem('g2', om.Group(), promotes_inputs=['x'])
        g2.add_subsystem('C2', om.ExecComp('y = 2*x', shape=1))
        g2.promotes('C2', inputs=['x'], src_indices=[1], src_shape=(2,))

        p.model.set_input_defaults('par.x', val=[7., -5.])

        p.setup()
        p.run_model()


@use_tempdirs
class Multipoint2TooManyProcs(SrcIndicesSerialMultipoint2):
    N_PROCS = 3


class SrcIndicesSerialMultipoint3(unittest.TestCase):

    def test_multipoint2(self):
        p = om.Problem()
        par = p.model.add_subsystem('par', om.ParallelGroup())

        g1 = par.add_subsystem('g1', om.Group(), promotes_inputs=['x'])
        g1.add_subsystem('C1', om.ExecComp('y = 3*x', shape=1))
        g1.promotes('C1', inputs=['x'], src_indices=[0], src_shape=(3,))

        g2 = par.add_subsystem('g2', om.Group(), promotes_inputs=['x'])
        g2.add_subsystem('C2', om.ExecComp('y = 2*x', shape=1))
        g2.promotes('C2', inputs=['x'], src_indices=[1], src_shape=(3,))

        g3 = par.add_subsystem('g3', om.Group(), promotes_inputs=['x'])
        g3.add_subsystem('C3', om.ExecComp('y = 5*x', shape=1))
        g3.promotes('C3', inputs=['x'], src_indices=[2], src_shape=(3,))

        p.model.set_input_defaults('par.x', val=[7., -5., 2.])

        p.setup()
        p.run_model()

        assert_check_totals(p.check_totals(of=['par.g1.C1.y', 'par.g2.C2.y', 'par.g3.C3.y'], wrt=['par.x']))


@use_tempdirs
class Multipoint3SameProcs(SrcIndicesSerialMultipoint3):
    N_PROCS = 3


@use_tempdirs
class Multipoint3TooManyProcs(SrcIndicesSerialMultipoint3):
    N_PROCS = 4

@use_tempdirs
class DoubleNestedParallelMultipointTestCase(unittest.TestCase):
    N_PROCS = 6

    def test_nested_pars(self):
        p = om.Problem()
        par = p.model.add_subsystem('par', om.ParallelGroup(), promotes_inputs=['x'])

        G1 = par.add_subsystem('G1', om.Group(), promotes_inputs=['x'])
        G1p = G1.add_subsystem('G1p', om.ParallelGroup(), promotes_inputs=['x'])
        G1p.add_subsystem('C1_1', om.ExecComp('y = 3.0*x'))
        G1p.add_subsystem('C1_2', om.ExecComp('y = -3.0*x'))
        G1p.promotes('C1_1', inputs=['x'], src_indices=[0])
        G1p.promotes('C1_2', inputs=['x'], src_indices=[0])

        G2 = par.add_subsystem('G2', om.Group(), promotes_inputs=['x'])
        G2p = G2.add_subsystem('G2p', om.ParallelGroup(), promotes_inputs=['x'])
        G2p.add_subsystem('C2_1', om.ExecComp('y = 5.0*x'))
        G2p.add_subsystem('C2_2', om.ExecComp('y = -5.0*x'))
        G2p.promotes('C2_1', inputs=['x'], src_indices=[1])
        G2p.promotes('C2_2', inputs=['x'], src_indices=[1])

        G3 = par.add_subsystem('G3', om.Group(), promotes_inputs=['x'], max_procs=1)  # no nested parallel group here
        G3s = G3.add_subsystem('G3s', om.Group(), promotes_inputs=['x'])
        G3s.add_subsystem('C3_1', om.ExecComp('y = 7.0*x'))
        G3s.add_subsystem('C3_2', om.ExecComp('y = -7.0*x'))
        G3s.promotes('C3_1', inputs=['x'], src_indices=[2])
        G3s.promotes('C3_2', inputs=['x'], src_indices=[2])

        par.set_input_defaults('x', val=[.5, 1.5, 2.5])

        p.setup()
        p.run_model()

        assert_check_totals(p.check_totals(of=['par.G1.G1p.C1_1.y', 'par.G1.G1p.C1_2.y',
                                               'par.G2.G2p.C2_1.y', 'par.G2.G2p.C2_2.y',
                                               'par.G3.G3s.C3_1.y', 'par.G3.G3s.C3_2.y'], wrt=['x'],
                                           show_only_incorrect=True))


class TestNestedInputDefaults(unittest.TestCase):
    N_PROCS = 3

    def test_nested(self):
        # this was failing as reported in github issue #2441
        # It passes if the two setup() calls don't raise any exceptions.

        class Grp(om.Group):
            def setup(self):
                self.add_subsystem("lin", om.ExecComp("y = x" ), promotes_inputs=["x"])
                self.add_subsystem("quad", om.ExecComp("z = x*x"), promotes_inputs=["x"])
                # self.set_input_defaults("x", 0.85)

        class Vec(om.Group):
            def initialize(self):
                self.options.declare("num", default=1)
            def setup(self):
                n = self.options["num"]

                self.set_input_defaults("x", np.full(n, 0.85))

                for node in range(n):
                    name = f"comp_{node}"
                    self.add_subsystem(name, Grp())
                    self.promotes(name, inputs=["x"], src_indices=[node])

        # This one seems to require that set_input_defaults (on line 8) is NOT called in Comp.setup()
        n = 3
        p = om.Problem(Vec(num=n))
        p.setup()

        # This one seems to require that set_input_defaults (on line 8) IS called in Comp.setup()
        p2 = om.Problem(Grp())
        p2.setup()


if __name__ == '__main__':
    unittest.main()



"""
A tool to make it easier to investigate coloring of jacobians with different sparsity structures.
"""

import numpy as np
from scipy.sparse import coo_matrix

import openmdao.api as om
from openmdao.utils.coloring import _compute_coloring
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.devtools.debug import compare_jacs


class TotJacBuilder(object):
    def __init__(self, nrows, ncols):
        self.J = np.zeros((nrows, ncols), dtype=bool)
        self.coloring = None

    def add_random_points(self, npoints):
        nrows, ncols = self.J.shape

        zro = self.J == False
        flat = self.J[zro].flatten()
        flat[:npoints] = True
        np.random.shuffle(flat)
        self.J[zro] = flat

    def add_row(self, idx, density=1.0):
        self.add_block(self.create_row(density=density), idx, 0)

    def add_col(self, idx, density=1.0):
        self.add_block(self.create_col(density=density), 0, idx)

    def create_row(self, density=1.0):
        return self.create_block((1, self.J.shape[1]), density=density)

    def create_col(self, density=1.0):
        return self.create_block((self.J.shape[0], 1), density=density)

    def create_block(self, shape, density=1.0):
        if density == 1.0:
            return np.ones(shape, dtype=bool)
        else:
            rows, cols = shape
            num = int((rows * cols) * density)
            vec = np.zeros(int(rows * cols), dtype=bool)
            vec[:num] = True
            np.random.shuffle(vec)
            return vec.reshape(shape)

    def add_block(self, block, start_row, start_col):
        rows, cols = block.shape
        self.J[start_row:start_row + rows, start_col:start_col + cols] = block

    def add_block_diag(self, shapes, start_row, start_col, density=1.0):
        row_idx = start_row
        col_idx = start_col

        for shape in shapes:
            self.add_block(self.create_block(shape, density=density), row_idx, col_idx)
            row_idx += shape[0]
            col_idx += shape[1]

    def color(self, mode='auto', fname=None, direct=True):
        self.coloring = _compute_coloring(self.J, mode, direct=direct)
        if self.coloring is not None and fname is not None:
            self.coloring.save(fname)
        return self.coloring

    def show(self):
        try:
            self.coloring.display_bokeh(show=True)
        except Exception:
            print("Bokeh not available, using ASCII display.")
            self.coloring.display_txt()

        maxdeg_fwd = np.max(np.count_nonzero(self.J, axis=1))
        maxdeg_rev = np.max(np.count_nonzero(self.J, axis=0))

        print("Shape:", self.J.shape)
        print("Density:", np.count_nonzero(self.J) / self.J.size)
        print("Max degree (fwd, rev):", maxdeg_fwd, maxdeg_rev)

        self.coloring.summary()

    def shuffle_rows(self):
        np.random.shuffle(self.J)

    def density_info(self):
        J = self.J
        density = np.count_nonzero(J) / J.size
        row_density = np.count_nonzero(J, axis=1) / J.shape[1]
        max_row_density = np.max(row_density)
        n_dense_rows = row_density[row_density == 1.0].size
        col_density = np.count_nonzero(J, axis=0) / J.shape[0]
        max_col_density = np.max(col_density)
        n_dense_cols = col_density[col_density == 1.0].size
        return density, max_row_density, n_dense_rows, max_col_density, n_dense_cols

    @staticmethod
    def make_blocks(num_blocks, min_shape, max_shape):
        shapes = []
        row_size = col_size = 0
        min_rows, min_cols = min_shape
        max_rows, max_cols = max_shape

        for b in range(num_blocks):
            nrows = np.random.randint(min_rows, max_rows + 1)
            ncols = np.random.randint(min_cols, max_cols + 1)
            shapes.append((nrows, ncols))
            row_size += nrows
            col_size += ncols

        return shapes, row_size, col_size

    @staticmethod
    def make_jac(n_dense_rows=0, row_density=1.0, n_dense_cols=0, col_density=1.0,
                 n_blocks=0, min_shape=(1,1), max_shape=(2,2), n_random_pts=0):
        if n_blocks > 0:
            shapes, nrows, ncols = TotJacBuilder.make_blocks(n_blocks, min_shape, max_shape)
            builder = TotJacBuilder(nrows + n_dense_rows, ncols + n_dense_cols)
            builder.add_block_diag(shapes, n_dense_rows, n_dense_cols)
        else:
            nrows, ncols = (100, 50)
            builder = TotJacBuilder(nrows, ncols)

        # dense rows
        for row in range(n_dense_rows):
            builder.add_row(row, density=row_density)

        # dense cols
        for col in range(n_dense_cols):
            builder.add_col(col, density=col_density)

        builder.add_random_points(n_random_pts)

        return builder

    @staticmethod
    def eisenstat(n):
        """
        Return a builder containing an Eisenstat's example Jacobian of size n+1 x n.

        Should be colorable with n/2 + 2 colors using bidirectional coloring.

        The columns in Eisenstat's example are pairwise structurally nonorthogonal,
        so a fwd directional coloring would require n groups.
        """
        assert n >= 6, "Eisenstat's example must have n >= 6."
        assert n % 2 == 0, "Eisenstat's example must have even 'n'."

        D1 = np.eye(n // 2, dtype=int)
        D2 = np.eye(n // 2, dtype=int)
        D3 = np.eye(n // 2, dtype=int)
        B = np.ones((n // 2, n // 2), dtype=int)
        idxs = np.arange(n // 2, dtype=int)
        B[idxs, idxs] = 0
        C = np.ones((1, n // 2), dtype=int)
        O = np.zeros((1, n // 2), dtype=int)

        A1 = np.hstack([D1, D2])
        A2 = np.vstack([np.hstack([C, O]), np.hstack([D3, B])])

        A = np.vstack([A1, A2])

        builder = TotJacBuilder(n + 1, n)
        builder.J[:, :] = A

        return builder


def rand_jac(minrows=(1, 10), mincols=(1, 10)):
    rnd = np.random.randint
    minr = rnd(*minrows)
    minc = rnd(*mincols)

    return  TotJacBuilder.make_jac(n_dense_rows=rnd(5), row_density=np.random.rand(),
                                   n_dense_cols=rnd(5), col_density=np.random.rand(),
                                   n_blocks=rnd(3,8),
                                   min_shape=(minr,minc),
                                   max_shape=(minr+rnd(10),minc+rnd(10)),
                                   n_random_pts=rnd(15))


class SparsityComp(om.ExplicitComponent):
    """
    A simple component that multiplies a sparse matrix by an input vector.

    The sparsity structure is defined by the 'sparsity' argument, and the data values are
    just the (index + 1) of the nonzeros in the sparsity structure.

    This component is used to test the coloring of the total jacobian.  A Problem is set up
    with a model containing only this component, and the total jacobian is computed with and
    without coloring.  The two jacobians are compared to ensure they are the same.

    Parameters
    ----------
    sparsity : ndarray or coo_matrix
        Sparsity structure to be tested.

    Attributes
    ----------
    sparsity : coo_matrix or ndarray
        Dense or sparse version of the sparsity structure.
    """
    def __init__(self, sparsity, **kwargs):
        super(SparsityComp, self).__init__(**kwargs)
        if isinstance(sparsity, np.ndarray):
            self.sparsity = coo_matrix(sparsity)
        else:
            self.sparsity = sparsity.tocoo()

        self.sparsity.data = np.arange(1, self.sparsity.data.size + 1)

    def setup(self):
        self.add_input('x', shape=self.sparsity.shape[1])
        self.add_output('y', shape=self.sparsity.shape[0])

        self.declare_partials('y', 'x', rows=self.sparsity.row, cols=self.sparsity.col)

    def compute(self, inputs, outputs):
        outputs['y'] = self.sparsity.dot(inputs['x'])

    def compute_partials(self, inputs, partials):
        partials['y', 'x'] = self.sparsity.data


def check_sparsity_tot_coloring(sparsity, direct=True, tolerance=1e-15, tol_type='rel', mode='auto'):
    """
    Check total derivatives of a top level SparsityComp with and without total coloring.

    Parameters
    ----------
    sparsity : ndarray or coo_matrix
        Sparsity structure to be tested.
    direct : bool
        If True, use the direct method to compute the column adjacency matrix when bidirectional
        coloring, else use the substitution method.
    """
    import sys

    # compute totals without coloring
    p = om.Problem()
    model = p.model
    model.add_subsystem('comp', SparsityComp(sparsity))
    model.add_design_var('comp.x')
    model.add_constraint('comp.y')
    p.driver = om.ScipyOptimizeDriver()
    p.driver.options['optimizer'] = 'SLSQP'
    p.setup(mode=mode)
    p.run_model()
    J = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

    # compute totals with coloring
    p = om.Problem()
    model = p.model
    model.add_subsystem('comp', SparsityComp(sparsity))
    model.add_design_var('comp.x')
    model.add_constraint('comp.y')
    p.driver = om.ScipyOptimizeDriver()
    p.driver.options['optimizer'] = 'SLSQP'
    p.driver.declare_coloring(direct=direct)
    p.setup(mode=mode)
    p.run_model()
    Jcolor = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

    # make sure totals match for both cases
    try:
        assert_near_equal(J, Jcolor, tolerance=tolerance, tol_type=tol_type)
    except Exception:
        diff = np.abs(J - Jcolor)
        mask = diff <= tolerance
        diff[mask] = 0.0
        if p.driver._coloring_info.coloring is not None and not p.driver._coloring_info._failed:
            print(p.driver._coloring_info.coloring, file=sys.stderr)

        with np.printoptions(linewidth=999, threshold=1000):
            print("Good J\n", J, file=sys.stderr)
            drows, dcols = np.nonzero(diff)
            print("J shape", J.shape, file=sys.stderr)
            print("J diff\n", list(zip(drows, dcols)), file=sys.stderr)
        raise


if __name__ == '__main__':
    import argparse
    import pickle

    parser = argparse.ArgumentParser()
    parser.add_argument("-e", "--eisenstat",
                        help="Build an Eisenstat's example matrix of size n+1 x n.",
                        action="store", type=int, default=-1, dest="eisenstat")
    parser.add_argument("-m", "--mode", type=str, dest="mode",
                        help="Direction of coloring (default is auto).",
                        default="auto")
    parser.add_argument('-s', '--save', dest="save", default=None,
                        help="Output file for jacobian so it can be reloaded and colored using"
                        " various methods for comparison.")
    parser.add_argument('-l', '--load', dest="load", default=None,
                        help="Input file for jacobian so it can be reloaded and colored using"
                        " various methods for comparison.")

    options = parser.parse_args()

    if options.load is not None:
        with open(options.load, "rb") as f:
            builder = pickle.load(f)
    elif options.eisenstat > -1:
        builder = TotJacBuilder.eisenstat(options.eisenstat)
    else:  # just do a random matrix
        builder = rand_jac()

    builder.color(options.mode)
    builder.show()

    if options.save is not None:
        with open(options.save, "wb") as f:
            pickle.dump(builder, f)



#!/usr/bin/env python
import argparse
import os
import subprocess
import pathlib

from copy_build_artifacts import copy_build_artifacts


_this_file = pathlib.Path(__file__).resolve()
REPO_ROOT = _this_file.parent
BOOK_DIR = pathlib.Path(REPO_ROOT, 'openmdao_book')


def build_book(book_dir=BOOK_DIR, clean=True):
    """
    Clean (if requested), build, and copy over necessary files for the JupyterBook to be created.
    Parameters
    ----------
    book_dir
    clean
    """
    save_cwd = os.getcwd()
    os.chdir(REPO_ROOT)
    if clean:
        subprocess.run(['jupyter-book', 'clean', book_dir])  # nosec: trusted input
    subprocess.run(['jupyter-book', 'build', '-W', book_dir])  # nosec: trusted input
    copy_build_artifacts(book_dir)
    os.chdir(save_cwd)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Build a JupyterBook and automatically copy over'
                                                 'necessary build artifacts')
    parser.add_argument('-c', '--clean', action='store_true',
                        help='Clean the old book out before building (default is False).')
    parser.add_argument('-b', '--book', action='store', default='openmdao_book',
                        help="The name of the book to be built (default is 'openmdao_book').")
    args = parser.parse_args()
    build_book(book_dir=args.book, clean=args.clean)





import os
import sys
import time
import argparse


def main():
    """
    A standalone program for testing ExternalCodeComp.

    Writes "test data" to the specified output file after an optional delay.
    Optionally writes the value of the environment variable "TEST_ENV_VAR"
    to the file.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("output_filename")
    parser.add_argument("-e", "--write_test_env_var",
                        help="Write the value of TEST_ENV_VAR to the file",
                        action="store_true", default=False)
    parser.add_argument("-d", "--delay", type=float,
                        help="time in seconds to delay")
    parser.add_argument("-r", "--return_code", type=int,
                        help="value to return as the return code", default=0)

    args = parser.parse_args()

    if args.delay:
        if args.delay < 0:
            raise ValueError('delay must be >= 0')
        time.sleep(args.delay)

    with open(args.output_filename, 'w') as out:
        out.write("test data\n")
        if args.write_test_env_var:
            out.write("%s\n" % os.environ['TEST_ENV_VAR'])

    return args.return_code


if __name__ == '__main__':
    sys.exit(main())
